// EBaseHashSetT.h: interface for the EBaseHashSetT class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EBASEHASHSETT_H__992AFCEA_3D03_11D3_BE1D_0040053AD648__INCLUDED_)
#define AFX_EBASEHASHSETT_H__992AFCEA_3D03_11D3_BE1D_0040053AD648__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000


template<class Key, class KeyArg>
class EBaseHashSetIteratorT;

/*! \class EBaseHashSetT<Key,KeyArg>
 * \include "EBaseHashSetT.h"
 * \brief Template use to store a set of element in hash table.
 */
template<class Key, class KeyArg>
class EBaseHashSetT
{
#ifdef _SUPPORT_TEMPLATE_FRIEND
  friend class EBaseHashSetIteratorT<Key,KeyArg>;
#endif

public:
  typedef EBaseHashSetIteratorT<Key,KeyArg> EIterator;

  /*! Constructor.
   * @param nSize initial size of the table.
   */
  EBaseHashSetT( SL nSize = 3 );

  /*! Destructor
   */
  virtual ~EBaseHashSetT();

  /*! Test if a element is present in the set.
   * @param key key associated to the item.
   * @return \c true if the element is present in the set.
   */
  bool operator[]( KeyArg key );

  /*! Lookup a item key using its key.
   * @param key key associated to the item.
   * @return pointeur on key if found. NULL if no item associated to key.
   */
  const Key *lookUp( KeyArg key ) const;

  /*! Add an item to the association table.
   * @param key associated to the item.
   * @param item item to add.
   */
  void add( KeyArg key );

  /*! Remove an item from the association table.
   * @param key key associated to the item to remove.
   */
  void remove( KeyArg key );

  /*! Remove all the item from the association table.
   */
  void removeAll();

  /*! Resize the table.
   * @param nSize new table size.
   */
  void resize( SL nSize );

  /*! Get the size of the table.
   * @return size of the table.
   */
  SL size() const;

  /*! Get the number of association in the table.
   * @return number of association in the table.
   */
  SL count() const;

  /*! Test if the table is empty.
   * @return true if there is no association in the table, false otherwise.
   */
  bool isEmpty() const;

#ifdef _SUPPORT_TEMPLATE_FRIEND
private:
#endif
  struct ELCAsso
  {
    ELCAsso *_pNext;
    Key _key;
  };

  EArray<ELCAsso *> _aAsso;

  SL _nNbAsso;

protected:
  virtual UL doGetKey( KeyArg key ) const =0;
  virtual bool doCompareKey( KeyArg key1, KeyArg key2 ) const =0;
  virtual void doFreeAsso( Key &key ) {}
};


// Includes:
// #####################Last change: 27/03/98 21h49mn#########################

#include "EBaseHashSetIteratorT.h"

// 
// #####################Last change: 27/03/98 21h49mn#########################

template<class Key, class KeyArg>
EBaseHashSetT<Key,KeyArg>::EBaseHashSetT( SL nSize )
{
  EASSERT( nSize > 0 );
  _nNbAsso = 0;
  resize( nSize );
}


template<class Key, class KeyArg>
EBaseHashSetT<Key,KeyArg>::~EBaseHashSetT()
{
  removeAll();
}


template<class Key, class KeyArg>
const Key *EBaseHashSetT<Key,KeyArg>::lookUp( KeyArg key ) const
{
  SL nIndex = doGetKey( key ) % _aAsso.count();
  ELCAsso *pAsso = _aAsso[ nIndex ];
  while ( pAsso )
  {
    if ( doCompareKey( pAsso->_key, key ) )
      return &pAsso->_key;
    pAsso = pAsso->_pNext;
  }
  return NULL;
}


template<class Key, class KeyArg>
bool EBaseHashSetT<Key,KeyArg>::operator[]( KeyArg key )
{
  return lookUp( key ) != NULL;
}


template<class Key, class KeyArg>
void EBaseHashSetT<Key,KeyArg>::add( KeyArg key )
{
  EASSERT( _aAsso.count() > 0 );
  EASSERT( !lookUp( key ) );    // Element must not be present in the set.

  SL nIndex = doGetKey( key ) % _aAsso.count();

  ELCAsso *pAsso = new ELCAsso();
  pAsso->_key = key;
  pAsso->_pNext = _aAsso[ nIndex ];
  _aAsso[ nIndex ] = pAsso;
  _nNbAsso++;
}


template<class Key, class KeyArg>
void EBaseHashSetT<Key,KeyArg>::remove( KeyArg key )
{
  EASSERT( _aAsso.count() > 0 );

  SL nIndex = doGetKey( key ) % _aAsso.count();
  ELCAsso *pAsso = _aAsso[ nIndex ];
  ELCAsso **pPrevAsso = &_aAsso[ nIndex ];
  while ( pAsso )
  {
    if ( doCompareKey( pAsso->_key, key ) )  // found !
    {
      *pPrevAsso = pAsso->_pNext;
      doFreeAsso( pAsso->_key );
      delete pAsso;
      _nNbAsso--;
      return;
    }

    pPrevAsso = &pAsso->_pNext;
    pAsso = pAsso->_pNext;
  }
  EASSERT( pAsso );
}


template<class Key, class KeyArg>
void EBaseHashSetT<Key,KeyArg>::resize( SL nSize )
{
  EDynArray<ELCAsso *> aAsso2;
  SL nCpt;
  ELCAsso *pNull = ENIL;

  aAsso2.resize( nSize );
  aAsso2.fill( &pNull );

  for ( nCpt = 0; nCpt < size(); ++nCpt )
  {
    ELCAsso *pAsso = _aAsso[ nCpt ];
    while ( pAsso )
    {
      ELCAsso *pNextAsso = pAsso->_pNext;
      SL nIndex = doGetKey( pAsso->_key ) % nSize;    // compute new index

      pAsso->_pNext = aAsso2[ nIndex ];              // add to head of lc
      aAsso2[ nIndex ] = pAsso;

      pAsso = pNextAsso;
    }
  }

  _aAsso = aAsso2;
}


template<class Key, class KeyArg>
inline SL EBaseHashSetT<Key,KeyArg>::size() const
{
  return _aAsso.count();
}


template<class Key, class KeyArg>
inline SL EBaseHashSetT<Key,KeyArg>::count() const
{
  return _nNbAsso;
}


template<class Key, class KeyArg>
void EBaseHashSetT<Key,KeyArg>::removeAll()
{
  SL nIndex;

  for ( nIndex = size() -1; nIndex >= 0; --nIndex )
  {
    ELCAsso *pAsso = _aAsso[ nIndex ];
    while ( pAsso )
    {
      ELCAsso *pNextAsso = pAsso->_pNext;
      doFreeAsso( pAsso->_key );
      delete pAsso;
      pAsso = pNextAsso;
    }
    _aAsso[ nIndex ] = ENIL;
  }
  _nNbAsso = 0;
}


template<class Key, class KeyArg>
bool 
EBaseHashSetT<Key,KeyArg>::isEmpty() const
{
  return count() == 0;
}



#endif // !defined(AFX_EBASEHASHSETT_H__992AFCEA_3D03_11D3_BE1D_0040053AD648__INCLUDED_)
