// EDListT.h: interface for the EDListT class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EDLISTT_H__3019D973_9829_11D3_AABD_E8D24E3A7463__INCLUDED_)
#define AFX_EDLISTT_H__3019D973_9829_11D3_AABD_E8D24E3A7463__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "EListCleanerT.h"

template<class Item, 
         class ItemArg=Item, 
         class Cleaner=EListCleanerNoneT<Item> >
class EDListT  
{
public:
  typedef Item TypeItem;
  typedef ItemArg TypeItemArg;
  typedef Cleaner TypeCleaner;
  typedef EDListT<Item,ItemArg,Cleaner> TypeList;

  /** Constructs a simple chained list.
   */
  EDListT();

  /** Destructor.
   *
   * Destruct the list and all its items.
   */
  virtual ~EDListT();

  /** Get the number of items in the list.
   */
  SL count() const;

  /** Test if the list is empty.
   * @return \c true if the list is empty, otherwise returns \c false.
   */
  bool isEmpty() const;

  /** Insert an item in the list.
   * @param item Item to insert.
   * @param nPosInsert 0 based position of insertion of the item. If <0 then
   *                   the position of insertion is \c count() + 1 + \a nPosInsert.
   *                   Therefore if \a nPosInsert is -1 the item will be inserted
   *                   at the end of the list.
   */
  void insert( ItemArg item, 
               SL nPosInsert );

  /** Insert an item before the first item of the list.
   * @param item Item to insert.
   */
  void prepend( ItemArg item );

  /** Insert an item after the last item of the list.
   * @param item Item to insert.
   */
  void append( ItemArg item );

  /** Remove an item from the list.
   * @param nPosRemove 0 based position of the item to remove. If <0 then
   *                   the position of the item is \c count() + \a nPosRemove.
   *                   Therefore if \a nPosRemove is -1 the last item will be 
   *                   removed.
   * @return \c true if an item was removed ( 0 <= \c PosRemove < count() ).
   */
  bool remove( SL nPosRemove );

  /** Remove an item from the list.
   * @param nPosRemove 0 based position of the item to remove. If <0 then
   *                   the position of the item is \c count() + \a nPosRemove.
   *                   Therefore if \a nPosRemove is -1 the last item will be 
   *                   removed.
   * @param nCount Number of items to remove. If <0 then the item are removed
   *               from the position resulting of \a nPosRemove value to \a 
   *               position nCount + count() included. If the ending position is less than
   *               the starting position, no item are removed. A value
   *               of -1 will remove all items after the position indicated by
   *               \a nPosRemove.
   * @return Number of items removed.
   */
  SL remove( SL nPosRemove,
             SL nCount );

  /** Destroy all the items of the list.
   */
  void reset();

  /** Remove the first item of the list.
   * @return \c true if the item was removed (the list is not empty).
   */
  bool removeFirst();

  /** Remove the last item of the list.
   * @return \c true if the item was removed (the list is not empty).
   */
  bool removeLast();

  /** Get a reference on the item at a given position.
   * @param nPos Position of the item. 0 <= \c nPos < count().
   * @return Reference on the item at the given position.
   */
  const Item &at( SL nPos ) const;

  /** Get a reference on the item at a given position.
   * @param nPos Position of the item. 0 <= \c nPos < count().
   * @return Reference on the item at the given position.
   */
  Item &at( SL nPos );

  /** Get a reference on the item at a given position.
   * @param nPos Position of the item. 0 <= \c nPos < count().
   * @return Reference on the item at the given position.
   */
  const Item &operator[]( SL nPos ) const;

  /** Get a reference on the item at a given position.
   * @param nPos Position of the item. 0 <= \c nPos < count().
   * @return Reference on the item at the given position.
   */
  Item &operator[]( SL nPos );

protected:
  /** \brief This structure represents a simple chained list node.
   */
  struct Node
  {
    /** Default constructor.
     *
     * Does nothing. Must only be used for copying an existing node.
     */
    Node() {}

    /** Constructs a node and initialize its item.
     *
     * The previous and next node are left uninitialized.
     * @param item Item that this node contains.
     */
    Node( ItemArg item ) : _item( item ) {}

    /** Constructs a node and initialize its item and next node.
     * @param item Item that this node contains.
     * @param pPrevNode Pointer on the previous node of the list. May be \c NULL.
     * @param pNextNode Pointer on the next node of the list. May be \c NULL.
     */
    Node( ItemArg item, 
          Node *pPrevNode,
          Node *pNextNode ) : _item( item ), 
                              _pPrevNode( pPrevNode ), 
                              _pNextNode( pNextNode ) {}


    /** Pointer on the previous node of the double chained list.
     */
    Node *_pPrevNode;

    /** Pointer on the next node of the simple chained list.
     */
    Node *_pNextNode;

    /** Item for this node.
     */
    Item _item;
  };

protected:
  /** Insert an item at a given position.
   * @param item Item to insert.
   * @param nPos 0 based insertion position. Must be valid.
   */
  void doInsert( ItemArg item, 
                 SL nPos );

  /** Remove an item at a given position.
   * @param nPos 0 based position of the item to remove.
   * @return \c true if the item was removed, that is \c nPos was a valid position
   *         in the list ( 0 <= \a nPos < count() ).
   */
  bool doRemove( SL nPos );


  /** Find a node at a given index.
   * @param nPos 0 based index of the node. 0 <= nPos < count(). 
   */
  Node *findNode( SL nPos ) const;

  /** Set the current node.
   * @param pNode Pointer on the current node. May be \c NULL.
   * @param nPos 0 based index of the current node. Not use if \a pNode is \c NULL.
   */
  void setCur( Node *pNode,
               SL nPos ) const;

private:
  /** Pointer on the first node of the list. May be \c NULL if the list is empty.
   */
  Node *_pListHead;

  /** Pointer on the last node of the list. May be \c NULL if the list is empty.
   */
  Node *_pListTail;

  /** Pointer on the last used node. May be \c NULL.
   */
  EMUTABLE Node *_pListCurrent;

  /** Position of the last used node in the list. 0 based.
   */
  EMUTABLE SL _nPosCurrent;

  /** Number of items in the list.
   */
  SL _nCount;

  /** Node cleaner.
   */
  Cleaner _cleaner;
};




// Template implementation class EDListT:
// --------------------------------------


template<class Item, class ItemArg, class Cleaner>
inline
EDListT<Item,ItemArg,Cleaner>::EDListT() :
    _pListHead( NULL ),
    _pListCurrent( NULL ),
    _nCount( 0 )
{
}


template<class Item, class ItemArg, class Cleaner>
inline
EDListT<Item,ItemArg,Cleaner>::~EDListT()
{
  reset();
}


template<class Item, class ItemArg, class Cleaner>
inline SL 
EDListT<Item,ItemArg,Cleaner>::count() const
{
  return _nCount;
}


template<class Item, class ItemArg, class Cleaner>
inline bool 
EDListT<Item,ItemArg,Cleaner>::isEmpty() const
{
  return count() == 0;
}


template<class Item, class ItemArg, class Cleaner>
inline void 
EDListT<Item,ItemArg,Cleaner>::insert( ItemArg item, 
                                       SL nPosInsert )
{
  if ( nPosInsert < 0 )
    nPosInsert += count() +1;

  EASSERT( nPosInsert >= 0 );
  EASSERT( nPosInsert <= count() );

  doInsert( item, nPosInsert );
}


template<class Item, class ItemArg, class Cleaner>
inline void 
EDListT<Item,ItemArg,Cleaner>::prepend( ItemArg item )
{
  doInsert( item, 0 );
}


template<class Item, class ItemArg, class Cleaner>
inline void 
EDListT<Item,ItemArg,Cleaner>::append( ItemArg item )
{
  doInsert( item, count() );
}


template<class Item, class ItemArg, class Cleaner>
inline bool 
EDListT<Item,ItemArg,Cleaner>::remove( SL nPosRemove )
{
  if ( nPosRemove < 0 )
    nPosRemove += count();

  return doRemove( nPosRemove );
}


template<class Item, class ItemArg, class Cleaner>
inline bool 
EDListT<Item,ItemArg,Cleaner>::removeFirst()
{
  return doRemove( 0 );
}


template<class Item, class ItemArg, class Cleaner>
inline bool 
EDListT<Item,ItemArg,Cleaner>::removeLast()
{
  return doRemove( count() -1 );
}


template<class Item, class ItemArg, class Cleaner>
SL 
EDListT<Item,ItemArg,Cleaner>::remove( SL nPosRemove,
                                       SL nCount )
{
  if ( nPosRemove < 0 )
    nPosRemove += count();

  if ( nCount < 0 )
    nCount = (count() + nCount) +1 - nPosRemove;

  SL nRemoved =0;
  while ( nCount-- > 0  &&
          doRemove( nPosRemove ) )
    ++nRemoved;

  return nRemoved;
}


template<class Item, class ItemArg, class Cleaner>
inline void 
EDListT<Item,ItemArg,Cleaner>::reset()
{
  remove( 0, -1 );
}


template<class Item, class ItemArg, class Cleaner>
inline const Item &
EDListT<Item,ItemArg,Cleaner>::at( SL nPos ) const
{
  return findNode( nPos )->_item;
}


template<class Item, class ItemArg, class Cleaner>
Item &
EDListT<Item,ItemArg,Cleaner>::at( SL nPos )
{
  return findNode( nPos )->_item;
}


template<class Item, class ItemArg, class Cleaner>
const Item &
EDListT<Item,ItemArg,Cleaner>::operator[]( SL nPos ) const
{
  return at( nPos );
}


template<class Item, class ItemArg, class Cleaner>
Item &
EDListT<Item,ItemArg,Cleaner>::operator[]( SL nPos )
{
  return at( nPos );
}


template<class Item, class ItemArg, class Cleaner>
void 
EDListT<Item,ItemArg,Cleaner>::doInsert( ItemArg item, 
                                         SL nPos )
{
  EASSERT( 0 <= nPos );
  EASSERT( nPos <= count() );

  Node *pNode;
  if ( nPos == 0)
  {
    pNode = new Node( item, NULL, _pListHead );
    _pListHead = pNode;
  }
  else
  {
    Node *pNodePrev = findNode( nPos -1 );
    pNode = new Node( item, pNodePrev, pNodePrev->_pNextNode );
    pNodePrev->_pNextNode = pNode;
  }

  if ( nPos == _nCount )
    _pListTail = pNode;

  ++_nCount;

  setCur( pNode, nPos );
}


template<class Item, class ItemArg, class Cleaner>
bool 
EDListT<Item,ItemArg,Cleaner>::doRemove( SL nPos )
{
  if ( nPos < 0  ||  nPos >= count() )
    return false;

  Node *pNode;
  if ( nPos == 0 )
  {
    pNode = _pListHead;
    _pListHead = pNode->_pNextNode;
    
    if ( nPos == _nPosCurrent )
      _pListCurrent = _pListHead;
  }
  else
  {
    Node *pPrevNode = findNode( nPos -1 );

    pNode = pPrevNode->_pNextNode;
    
    Node *pNextNode = pNode->_pNextNode;
    pPrevNode->_pNextNode = pNextNode;
    pNextNode->_pPrevNode = pPrevNode;

    if ( nPos == _nPosCurrent )
    {
      _pListCurrent = pPrevNode;
      --_nPosCurrent;
    }
  }

  if ( nPos == _nCount-1 )
    _pListTail = pNode->_pPrevNode;

  _cleaner.destroy( pNode->_item );
  delete pNode;
  --_nCount;
  return true;
}


template<class Item, class ItemArg, class Cleaner>
EDListT<Item,ItemArg,Cleaner>::Node *
EDListT<Item,ItemArg,Cleaner>::findNode( SL nPos ) const
{
  EASSERT( 0 <= nPos );
  EASSERT( nPos < count() );

  Node *pNode;
  SL nDistFirst = nPos;
  SL nDistLast = count() -1 -nPos;
  SL nDistCurrent = _pListCurrent ? eabs( _nPosCurrent - nPos ) : count();

  SL nStep;
  bool bForward;
  bool bIsCurrent;

  if ( nDistFirst <= nDistCurrent )
  {
    if ( nDistFirst <= nDistLast )    // proceed from the head of the list
    {
      pNode = _pListHead;
      bForward = true;
      nStep = nPos;
    }
    else  // nDistLast < nDistFirst :    proceed from the tail of the list 
    {
      pNode = _pListTail;
      bForward = false;
      nStep = nDistLast;
    }
    bIsCurrent = false;
  }
  else // nDistCurrent < nDistFirst
  {
    if ( nDistCurrent < nDistLast )   // proceed from the current position
    {
      nStep = eabs( nPos - _nPosCurrent );
      pNode = _pListCurrent;
      bForward = ( nPos > _nPosCurrent );
      bIsCurrent = true;
    }
    else  // nDistLast < nDistCurrent // proceed from the tail of the list
    {
      pNode = _pListTail;
      bForward = false;
      nStep = nDistLast;
      bIsCurrent = false;
    }
  }

  if ( bForward )
  {
    while ( nStep-- )
      pNode = pNode->_pNextNode;
  }
  else
  {
    while ( nStep-- )
      pNode = pNode->_pPrevNode;
  }

  if ( bIsCurrent )
  {
    _pListCurrent = pNode;
    _nPosCurrent = nPos;
  }

  return pNode;
}


template<class Item, class ItemArg, class Cleaner>
inline void 
EDListT<Item,ItemArg,Cleaner>::setCur( Node *pNode,
                                       SL nPos ) const
{
  EASSERT( pNode  &&  nPos >= 0 );
  EMUTABLE_DO(TypeList)->_pListCurrent = pNode;
  EMUTABLE_DO(TypeList)->_nPosCurrent = nPos;
}


#endif // !defined(AFX_EDLISTT_H__3019D973_9829_11D3_AABD_E8D24E3A7463__INCLUDED_)
