// EFastBasicStack.h: interface for the EFastBasicStack class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_EFASTBASICSTACK_H__FEA0CFA9_91FD_11D2_AF1C_AED26EBCA961__INCLUDED_)
#define AFX_EFASTBASICSTACK_H__FEA0CFA9_91FD_11D2_AF1C_AED26EBCA961__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "EBasicStack.h"

/* \brief This class represents a generic stack for simple type.
 *
 * @param TElm Type of elements stored in the stack.
 *
 * For pushing, argument is a reference on the element to push.
 * 
 * Element must have a copy operator, must not used constructor & destructor.
 */
template<class TElm>
class EFastBasicStack : public EBasicStack<TElm>
{
  public:
    /*! Contructs a Fast Stack.
     * @param defaultSize number of element the stack can handle first without
     *        any resizing.
     */
    EFastBasicStack( SL defaultSize =0 );

    /*! Destructor.
     */
    virtual ~EFastBasicStack() {}

    /*! Tests if the stack is empty.
     * @return \c true if the stack is empty.
     */
    virtual bool isEmpty() const;

    /*! Push an element on the stack.
     * @param elm Element to push on the stack.
     */
    virtual void push( TElm elm );

    /*! Pop an element from the stack.
     * Pop an element from the stack. The Stack must not be empty.
     * @return Poped element.
     * @exception EXCStackEmpty raise if the stack is empty.
     */
    virtual TElm pop();

    /*! Get the top element of the stack.
     * 
     * @return Reference on the top element of the stack. The stack must not be empty.
     * @exception EXCStackEmpty raise if the stack is empty.
     */
    virtual TElm &top();

    /*! Clear the stack.
     *
     * Remove all elements from the stack.
     */
    virtual void clear();

    /*! Minimize the size of the stack.
     *
     * Resize the dynamic array to the number of elements in the stack.
     */
    void shrink();

private:
  EDynArray<TElm> _aStack;
};




template<class TElm>
EFastBasicStack<TElm>::EFastBasicStack( SL defaultSize ) :
  _aStack( defaultSize )
{
}


template<class TElm>
inline bool 
EFastBasicStack<TElm>::isEmpty() const
{
  return _aStack.count() == 0;
}


template<class TElm>
inline void 
EFastBasicStack<TElm>::push( TElm elm )
{
  _aStack.append( elm );
}


template<class TElm>
inline TElm
EFastBasicStack<TElm>::pop()
{
  SL nIdx = _aStack.count() -1;
  if ( nIdx < 0 )
    ETHROW( EXCStackEmpty() );

  TElm elm = _aStack[ nIdx ];
  _aStack.resize( nIdx );
  
  return elm;
}


template<class TElm>
inline TElm &
EFastBasicStack<TElm>::top()
{
  SL nIdx = _aStack.count() -1;
  if ( nIdx < 0 )
    ETHROW( EXCStackEmpty() );

  return _aStack[ nIdx ];
}


template<class TElm>
void 
EFastBasicStack<TElm>::clear()
{
  _aStack.resize( 0 );
}


template<class TElm>
inline void 
EFastBasicStack<TElm>::shrink()
{
  _aStack.shrink();
}



#endif // !defined(AFX_EFASTBASICSTACK_H__FEA0CFA9_91FD_11D2_AF1C_AED26EBCA961__INCLUDED_)
