// ###########################################################################
// File: EHASHVAL.HPP
// (c)Copyright 03/1998, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// ###########################################################################


#ifndef EHASHVAL_HPP_INCLUDED


/* History:
 * 13/09/1998: BugFix: lookUp() const; Non finite recursity ! => copy/paste non const method.
 * 19/10/1998: BugFix: remove(), syntax error !
 */

// Includes:
// #####################Last change: 24/07/98 21h49mn#########################
#include "EDynArray.h"

// Constantes:
// #####################Last change: 24/07/98 21h49mn#########################


// Macros:
// #####################Last change: 24/07/98 21h49mn#########################


// Types:
// #####################Last change: 24/07/98 21h49mn#########################

template<class Key, class Item, class KeyArg, class ItemArg>
class EHashValIterator;


/* Hash Table.
 * Key Type of the key, must have a numeric (of type UL) value and the operator ==.
 * Item Type of the item associated with a key, for storage & return value.
 * KeyArg and ItemArg are the as Key & Item excepts that they are used when
 * passing paramaters (usually it is a reference on Key and Item type).
 *
 * To use this class you must derivate it and overide the virtual method
 * getKey() that compute a integer value based on the key and compareKey()
 * that compare two keys.
 *
 * Exemple of use:
 * Here we want to use an hash table to compute frequency of words in a text.
 * Therefore the key is the word and the item associated with the key is an
 * integer.
 *
 * Here is the class declaration:
 *
 *   Notes: Here the type for storing word and frequency are basic type,
 * therefore they no need to use reference when passing type as parameter.
 *
 *
 *   typedef EHashVal<UL,char *,UL,char *,UL> HashWordByFrq 
 *
 *   You can now add words:
 *
 *   HashWordByFrq hash;
 *
 *   if ( !hash.lookUp( 1 ) )  // the word is not in the table !
 *     hash.add( 1, "Berto" );         // Add the word to table & frq = 1
 *
 * If you override freeAsso(), you must also override the destructor, and call
 * removeAll() on destruction, otherwise the overrided freeAsso() will not be
 * called.
 */


template<class Key, class Item, class KeyArg, class ItemArg>
class EHashVal
{
#ifdef _SUPPORT_TEMPLATE_FRIEND
  friend class EHashValIterator<Key,Item,KeyArg,ItemArg>;
#endif

public:
  /// Alias for iterator.
  typedef EHashValIterator<Key,Item,KeyArg,ItemArg> EIterator;

  /** Constructor.
   * @param nSize initial size of the table.
   */
  EHashVal( SL nSize = 3 );

  // Destructor
  virtual ~EHashVal();

  /** Lookup a item using its key.
   * @param key key associated to the item.
   * @return reference on item.
   * */
  Item& operator[]( KeyArg key );

  /** Lookup a item using its key.
   * @param key key associated to the item.
   * @return reference on item.
   * */
//  const Item& operator[]( const KeyArg key ) const;

  /** Lookup a item using its key.
   * @param key key associated to the item.
   * @return pointeur on item if found. NULL if no item associated to key.
   * */
  Item *lookUp( KeyArg key );

  /** Lookup a item using its key.
   * @param key key associated to the item.
   * @return pointeur on item if found. NULL if no item associated to key.
   * */
  const Item *lookUp( KeyArg key ) const;

  /** Add an item to the association table.
   * @param key associated to the item.
   * @param item item to add.
   */
  void add( KeyArg key, ItemArg item );

  /** Remove an item from the association table.
   * @param key key associated to the item to remove.
   */
  void remove( KeyArg key );

  /// Remove all the item from the association table.
  void removeAll();

  /** Resize the table.
   * @param nSize new table size.
   */
  void resize( SL nSize );

  /** Get the size of the table.
   * @return size of the table.
   */
  SL size() const;

  /** Get the number of association in the table.
   * @return number of association in the table.
   */
  SL count() const;

  /** Test if the table is empty.
   * @return true if there is no association in the table, false otherwise.
   */
  bool isEmpty() const;

#ifdef _SUPPORT_TEMPLATE_FRIEND
private:
#endif
  struct ELCAsso
  {
    ELCAsso *_pNext;
    Key _key;
    Item _item;
  };

  EPtrDynArrayT<ELCAsso *> _aAsso;

  SL _nNbAsso;
  
protected:
  /** Free memory use by a key and an item (in case they use indirect data).
   * For example if key is a "char *" then you need to do "free( key )".
   * @param key  key to free
   * @param item item item to free.
   */
  virtual void freeAsso( Key &key, Item &item ) {}
};


// Template Implementation:
// #####################Last change: 24/07/98 21h49mn#########################

/** Constructs an hash table object.
 * @param nSize size of the hash table. To avoid collision, the rules of the
 *        thumb is that nSize should be 1.3 times the number of associations.
 *        Giving large size does not use lot of memory (only 1 (void *) ptr by
 *        unit).
 */
template<class Key, class Item, class KeyArg, class ItemArg>
EHashVal<Key,Item,KeyArg,ItemArg>::EHashVal( SL nSize )
{
  EASSERT( nSize > 0 );
  _nNbAsso = 0;
  resize( nSize );
}
// Destructor
template<class Key, class Item, class KeyArg, class ItemArg>
EHashVal<Key,Item,KeyArg,ItemArg>::~EHashVal()
{
  removeAll();
}

/** Lookup a item using its key.
  * @param key key associated to the item.
  * @return pointeur on item if found. NULL if no item associated to key.
  * */
template<class Key, class Item, class KeyArg, class ItemArg>
Item *EHashVal<Key,Item,KeyArg,ItemArg>::lookUp( KeyArg key )
{
  SL nIndex = (UL)(key) % _aAsso.count();
  ELCAsso *pAsso = _aAsso[ nIndex ];
  while ( pAsso )
  {
    if ( pAsso->_key == key )
      return &pAsso->_item;
    pAsso = pAsso->_pNext;
  }
  return NULL;
}


/** Lookup a item using its key.
  * @param key key associated to the item.
  * @return pointeur on item if found. NULL if no item associated to key.
  */
template<class Key, class Item, class KeyArg, class ItemArg>
const Item *EHashVal<Key,Item,KeyArg,ItemArg>::lookUp( KeyArg key ) const
{
  SL nIndex = (UL)(key) % _aAsso.count();
  ELCAsso *pAsso = _aAsso[ nIndex ];
  while ( pAsso )
  {
    if ( pAsso->_key == key )
      return &pAsso->_item;
    pAsso = pAsso->_pNext;
  }
  return NULL;
}


/** Lookup a item using its key.
  * @param key key associated to the item.
  * @return reference on item.
  * */
template<class Key, class Item, class KeyArg, class ItemArg>
Item& EHashVal<Key,Item,KeyArg,ItemArg>::operator[]( KeyArg key )
{
  Item *pItem = lookUp( key );
  EASSERT( pItem );
  return *pItem;
}

/** Lookup a item using its key.
  * @param key key associated to the item.
  * @return reference on item.
  * */
/*
template<class Key, class Item, class KeyArg, class ItemArg>
const Item& EHashVal<Key,Item,KeyArg,ItemArg>::operator[]( const Key key ) const
{
  const Item *pItem = lookUp( key );
  EASSERT( pItem );
  return *pItem;
}
*/
/** Add an item to the association table.
  * @param key associated to the item.
  * @param item item to add.
  */
template<class Key, class Item, class KeyArg, class ItemArg>
void EHashVal<Key,Item,KeyArg,ItemArg>::add( KeyArg key,
                                          ItemArg item )
{
  EASSERT( _aAsso.count() > 0 );

  SL nIndex = (UL)(key) % _aAsso.count();

  ELCAsso *pAsso = new ELCAsso();
  pAsso->_key = key;
  pAsso->_item = item;
  pAsso->_pNext = _aAsso[ nIndex ];
  _aAsso[ nIndex ] = pAsso;
  _nNbAsso++;
}


/** Remove an item from the association table.
  * @param key key associated to the item to remove.
  */
template<class Key, class Item, class KeyArg, class ItemArg>
void EHashVal<Key,Item,KeyArg,ItemArg>::remove( KeyArg key )
{
  EASSERT( _aAsso.count() > 0 );

  SL nIndex = (UL)(key) % _aAsso.count();
  ELCAsso *pAsso = _aAsso[ nIndex ];
  ELCAsso **pPrevAsso = &_aAsso[ nIndex ];
  while ( pAsso )
  {
    if ( pAsso->_key == key )  // found !
    {
      *pPrevAsso = pAsso->_pNext;
      freeAsso( pAsso->_key, pAsso->_item );
      delete pAsso;
      _nNbAsso--;
      return;
    }

    pPrevAsso = &pAsso->_pNext;
    pAsso = pAsso->_pNext;
  }
  EASSERT( pAsso );
}


/** Resize the table.
  * @param nSize new table size.
  */
template<class Key, class Item, class KeyArg, class ItemArg>
void EHashVal<Key,Item,KeyArg,ItemArg>::resize( SL nSize )
{
  EPtrDynArrayT<ELCAsso *> aAsso2;
  SL nCpt;
  ELCAsso *pNull = ENIL;

  _aAsso.setOwnership( false );

  aAsso2.resize( nSize );
  aAsso2.setOwnership( false );
//  aAsso2.fill( &pNull );

  for ( nCpt = 0; nCpt < size(); ++nCpt )
  {
    ELCAsso *pAsso = _aAsso[ nCpt ];
    while ( pAsso )
    {
      ELCAsso *pNextAsso = pAsso->_pNext;
      SL nIndex = (UL)( pAsso->_key ) % nSize;    // compute new index

      pAsso->_pNext = aAsso2[ nIndex ];              // add to head of lc
      aAsso2[ nIndex ] = pAsso;

      pAsso = pNextAsso;
    }
  }

  _aAsso = aAsso2;
  _aAsso.setOwnership( true );
}


template<class Key, class Item, class KeyArg, class ItemArg>
inline SL EHashVal<Key,Item,KeyArg,ItemArg>::size() const
{
  return _aAsso.count();
}


template<class Key, class Item, class KeyArg, class ItemArg>
inline SL EHashVal<Key,Item,KeyArg,ItemArg>::count() const
{
  return _nNbAsso;
}

template<class Key, class Item, class KeyArg, class ItemArg>
void EHashVal<Key,Item,KeyArg,ItemArg>::removeAll()
{
  SL nIndex;

  for ( nIndex = size() -1; nIndex >= 0; --nIndex )
  {
    ELCAsso *pAsso = _aAsso[ nIndex ];
    while ( pAsso )
    {
      ELCAsso *pNextAsso = pAsso->_pNext;
      freeAsso( pAsso->_key, pAsso->_item );
      delete pAsso;
      pAsso = pNextAsso;
    }
    _aAsso[ nIndex ] = ENIL;
  }
  _nNbAsso = 0;
}

template<class Key, class Item, class KeyArg, class ItemArg>
bool 
EHashVal<Key,Item,KeyArg,ItemArg>::isEmpty() const
{
  return count() == 0;
}


// Includes:
// #####################Last change: 28/03/98 20h01mn#########################
#include "EHValIte.hpp"


#define EHASHVAL_HPP_INCLUDED
#endif


// ############################ EOF EHASHVAL.HPP ################################
// ############################ EOF EHASHVAL.HPP ################################
// ############################ EOF EHASHVAL.HPP ################################
