// ###########################################################################
// File: EFASTDEV.HPP
// (c)Copyright 03/1998, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// Version 1.0
// ###########################################################################


#ifndef EFASTDEV_HPP_INCLUDED

// Includes:
// #####################Last change: 28/03/98 23h10mn#########################
#include "eiodevic.hpp"

// Constantes:
// #####################Last change: 28/03/98 23h10mn#########################


// Macros:
// #####################Last change: 28/03/98 23h10mn#########################


// Types:
// #####################Last change: 28/03/98 23h10mn#########################

class EFastDevice : public EIODevice
{
public:
  /** Constructs a fast device for an i/o device.
   * @param pDevice pDevice the fast device read/write info from/to.
   * @param nCacheSize cache size.
   */
  EFastDevice( EIODevice *pDevice = ENIL, SL nCacheSize =32768);

  /** Destructor.
   * Close device if open.
   */
  virtual ~EFastDevice();

  /** Open the file.
    * Open the file of name given on construction or with \l setName() using
    * mode \a _mode.
    * @param _mode mode used to open the file.
    * @see EIODevice::flags()
    * @exception EXCIODevice raised on error
    */
  void open( int _mode );

  // Close the file.
  void close();

  // Flush the file.
  void flush();

  // Return the size of the file
  SL size() const;

  /** Return the index of the file.
    * @return index of the file.
    */
  SL at() const;

  /** Set the index of the file.
    * If \a pos is beyond end of file, then the file is expanded.
    * The new region is initialised with 0.
    * @param pos position of the index in the file.
    */
  void at( SL pos );

  /** Test if the index is at the end of the file.
    * @return true if the index is at the end of the file.
    */
  bool atEnd() const;

  /** Read a block of byte from the device.
    * Read \a len Byte from the device and store them in the buffer \a data.
    * @param data buffer where data are stored.
    * @param len number of bytes of data to read.
    * @return number of byte of data actually read.
    * @exception EXCIODevice is case of grave error.
    */
  SL readBlock( void *data, SL len);

  /** Write a block of byte on the device.
    * Write \a len Byte of buffer \a data on the device.
    * @param data buffer where data are read.
    * @param len number of bytes of data to write.
    * @exception EXCIODevice is case of grave error.
    */
  void writeBlock( const void *data, SL len );

  /** Read a byte from the stream.
    * @return byte read.
    * @exception EIODevice on error
    */
  UB getByte();

  /** Write a Byte on the stream.
    * @param c byte to write.
    * @exception EIODevice on error.
    */
  void putByte( UB c );

  /** Push-back a byte on the stream.
    * @param c byte to push-back.
    * @exception EIODevice on error.
    */
  void ungetByte( UB c );

  /// Set access as exclusive read mode.
  void readMode();

  /// Set access as exclusive read mode.
  void writeMode();

  /** Attach a device to the fast file.
    * @param pDevice pDevice the fast device read/write info from/to.
    */
  void setDevice( EIODevice *pDevice );

  /** Get the attached device.
   * @return attached device.
   */
  EIODevice *device() const;

private:
  /** Shift the cache buffer.
   * @param nShift number of byte to shift. If nShift > 0 then shift cache
   *        toward the right ( [0] -> [nShift] ). If nShift < 0 then shift
   *        cache toward the left ( [nShift] -> [0] ).
   */
  void shiftCache( SL nShift );

  /** Step cache forward by _nCacheStep byte for reading.
   * @exception EXCIODevice.
   */
  void stepCacheForwardRead();

  /** Step cache forward by _nCacheStep byte for writing.
   * @exception EXCIODevice.
   */
  void stepCacheForwardWrite();

  /** Prepare cache for reading.
   * @param nPos position in file.
   */
  void startRead( SL nPos );

  /** Prepare cache for writing.
   * @param nPos position in file.
   */
  void startWrite( SL nPos );


  /// Associated device.
  EIODevice *_pDevice;

  /// Position of the first byte of the cache in the file.
  SL _nCachePos;

  /// Number of byte available in the cache (when reaching the eof ).
  SL _nCacheSize;

  /// How many byte of the original buffer are left when step ahead in the file.
  SL _nCacheStep;

  /// The cache.
  UB *_pCache;

  /// size of the allocated buffer for the cache.
  SL _nCacheMaxSize;

  /// Pointer to the end of the cache ( =_pCache + _nCacheSize )
  UB *_pEndCache;

  /// Current position in the cache.
  UB *_pCurByteInCache;

  /// Current acces mode.
  bool _bAccessReadMode;
};

// inlines:
// #####################Last change: 28/03/98 23h10mn#########################

inline UB EFastDevice::getByte()
{
  EASSERT( _bAccessReadMode );
  if ( _pCurByteInCache >= _pEndCache )
    stepCacheForwardRead();
  return *_pCurByteInCache++;
}

inline void EFastDevice::putByte( UB c )
{
  EASSERT( !_bAccessReadMode );
  if ( _pCurByteInCache >= _pEndCache )
    stepCacheForwardWrite();
  *_pCurByteInCache++ = c;
}

inline EIODevice *EFastDevice::device() const
{
  return _pDevice;
}

// Prototypes:
// #####################Last change: 28/03/98 23h10mn#########################



#define EFASTDEV_HPP_INCLUDED
#endif


// ########################### EOF EFASTDEV.HPP ##############################
// ########################### EOF EFASTDEV.HPP ##############################
// ########################### EOF EFASTDEV.HPP ##############################

