// ###########################################################################
// File: EFILE.HPP
// (c)Copyright 08/1997, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// version 0.2
// ###########################################################################

#ifndef EFILE_HPP_INCLUDED



#include <stdio.h>
#include "eiodevic.hpp"

class EFile : public EIODevice
{
public:
    // Default Constructor
    EFile();

    /** Constructs a file with a given name.
     * Constructs a file with a file name \a name.
     * @param _name name of the file
     */
    EFile( const char *_name );

    // Destructor
    virtual ~EFile();

    // set the name of the file.
    void setName( const char *name );

    /** Test if the file exists.
     */
    bool exists() const;

    /** Test if a file exists.
     * @param _name name of the file to test.
     */
    static bool exists( const char *fileName );

    /** Open the file.
     * Open the file of name given on construction or with \l setName() using
     * mode \a _mode.
     * @param _mode mode used to open the file.
     * @see EIODevice::flags()
     * @exception EXCIODevice raised on error
     */
    void open( int _mode );

    // Close the file.
    void close();

    // Flush the file.
    void flush();

    // Return the size of the file
    SL size() const;

    /** Return the index of the file.
     * @return index of the file.
     */
    SL at() const;

    /** Set the index of the file.
     * If \a pos is beyond end of file, then the file is expanded.
     * The new region is initialised with 0.
     * @param pos position of the index in the file.
     */
    void at( SL pos );

    /** Test if the index is at the end of the file.
     * @return true if the index is at the end of the file.
     */
    bool atEnd() const;

    /** Read a block of byte from the device.
     * Read \a len Byte from the device and store them in the buffer \a data.
     * @param data buffer where data are stored.
     * @param len number of bytes of data to read.
     * @return number of byte of data actually read.
     * @exception EXCIODevice is case of grave error.
     */
    SL readBlock( void *data, SL len);

    /** Write a block of byte on the device.
     * Write \a len Byte of buffer \a data on the device.
     * @param data buffer where data are read.
     * @param len number of bytes of data to write.
     * @exception EXCIODevice is case of grave error.
     */
    void writeBlock( const void *data, SL len );

    /** Read a byte from the stream.
     * @return byte read.
     * @exception EIODevice on error
     */
    UB getByte();

    /** Write a Byte on the stream.
     * @param c byte to write.
     * @exception EIODevice on error.
     */
    void putByte( UB c );

    /** Push-back a byte on the stream.
     * @param c byte to push-back.
     * @exception EIODevice on error.
     */
    void ungetByte( UB c );

private:
    /// Initialisation
    void init();

    /// stream
    FILE *mFile;

    /// filename
    char *mFilename;

    /// length of the file
    SL mLength;
};

// Inlines:
// ###########################################################################


/** Return the index of the file.
  * @return index of the file.
  */
inline SL EFile::at() const
{
  EASSERT( isOpen() );
  return mIndex;
}

/** Test if the index is at the end of the file.
  * @return true if the index is at the end of the file.
  */
inline bool EFile::atEnd() const
{
  EASSERT( isOpen() );
  return at() >= size() ? true : false;
}

#define EFILE_HPP_INCLUDED
#endif

// ############################ EOF EFILE.HPP ################################
// ############################ EOF EFILE.HPP ################################
// ############################ EOF EFILE.HPP ################################

