// ###########################################################################
// File: EIODEVIC.HPP
// (c)Copyright 08/1997, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// version 0.2
// ###########################################################################

#ifndef EIODEVIC_HPP_INCLUDED

#include "exciodev.hpp"

/* HST:
 * 08/10/98: - Added shared mode defintion, though it is not implemented...
 */


#define EIO_ReadOnly       0x0001
#define EIO_WriteOnly      0x0002
#define EIO_ReadWrite      0x0004
#define EIO_Append         0x0008
#define EIO_Truncate       0x0010
#define EIO_ModeMask       0x00ff

#define EIO_IsOpen         0x0100
#define EIO_StatusMask     0x0f00

#define EIO_ShareDenyRead  0x1000
#define EIO_ShareDenyWrite 0x2000
#define EIO_ShareExclusive 0x4000
#define EIO_ShareMask      0x7000 

class EIODevice
{
public:
    /// Default Constructor
    EIODevice();

    /// Destructor
    virtual ~EIODevice();

    /** Return the access mode.
     * Return the access mode to the I/O device.
     * @return access mode.
     */
    int accessMode() const;

    /** Return the status
     * @return status.
     */
    int status() const;

    /** Test if the I/O Device is readable.
     * @return true if it is readable.
     */
    bool isReadable() const;

    /** Test if the I/O Device is writable.
     * @return true if it is writable.
     */
    bool isWritable() const;

    /** Test if the I/O Device is readable and writable.
     * @return true if it is readable and writable.
     */
    bool isReadWrite() const;

    /** Test if the I/O Device is readable.
     * @return true if it is readable.
     */
    bool isOpen() const;


    /** Open the device.
     * Open the device using mode \a _mode.
     * @param _mode mode use to open the device
     */
    virtual void open( int _mode ) =0;

    /** Close the device.
     */
    virtual void close() =0;

    /** Flush the device.
     */
    virtual void flush() =0;

    /** Return the size of the device.
     * @return size of the device.
     */
    virtual SL size() const =0;

    /** Return the index of the device.
     * @return index of the device.
     * @exception EXCIODevice on error.
     */
    virtual SL at() const =0;

    /** Set the position of the device.
     * Set the position of the device (index) to \a pos. If \a pos if beyond
     * the size of the device, then the device is extended. The extension
     * region is initialised with 0.
     * @param pos new position of the index.
     * @exception EXCIODevice on error.
     */
    virtual void at( SL pos) =0;

    /** Test if the device is at the end.
     * @return true if the device is at the end, false otherwise.
     */
    virtual bool atEnd() const =0;

    /** Read a block of byte from the device.
     * Read \a len Byte from the device and store them in the buffer \a data.
     * @param data buffer where data are stored.
     * @param len number of bytes of data to read.
     * @return number of byte of data actually read.
     * @exception EXCIODevice is case of grave error.
     */
    virtual SL readBlock( void *data, SL len) =0;

    /** Write a block of byte on the device.
     * Write \a len Byte of buffer \a data on the device.
     * @param data buffer where data are read.
     * @param len number of bytes of data to write.
     * @exception EXCIODevice is case of grave error.
     */
    virtual void writeBlock( const void *data, SL len ) =0;

    /** Read a byte from the stream.
     * @return byte read.
     * @exception EIODevice on error
     */
    virtual UB getByte() =0;

    /** Write a Byte on the stream.
     * @param c byte to write.
     * @exception EIODevice on error.
     */
    virtual void putByte( UB c ) =0;

    /** Push-back a byte on the stream.
     * @param c byte to push-back.
     * @exception EIODevice on error.
     */
    virtual void ungetByte( UB c ) =0;

protected:
    /** Set access mode.
     * Set access mode for the device.
     * @param mode access mode.
     */
    void setMode( int mode );

    /** Set a status.
     * Set a status bit on.
     * @param state State to set.
     */
    void setStatus( int state );

    /** Unset a status.
     * Unset a status bit on.
     * @param state State to unset.
     */
    void unsetStatus( int state );

    /// reset access mode & status
    void resetInfo();

    /// Position in file.
    SL mIndex;

private:
    EIODevice( const EIODevice& ) {}; // Disable copy & = operators.
    EIODevice& operator =( const EIODevice& ) { return *this; }

    /// Information about mode...
    int info;
};

/// Default Constructor
inline EIODevice::EIODevice()
{
    mIndex =0;
    resetInfo();
}

/// Destructor
inline EIODevice::~EIODevice()
{
}

/** Return the access mode.
 * Return the access mode to the I/O device.
 * @return access mode.
 */
inline int EIODevice::accessMode() const
{
  return info & EIO_ModeMask;
}

/** Return the status
 * @return status.
 */
inline int EIODevice::status() const
{
  return info & EIO_StatusMask;
}

/** Test if the I/O Device is readable.
 * @return true if it is readable.
 */
inline bool EIODevice::isReadable() const
{
  return (info & (EIO_ReadOnly | EIO_ReadWrite) ) != 0;
}


/** Test if the I/O Device is readable.
 * @return true if it is readable.
 */
inline bool EIODevice::isWritable() const
{
  return (info & (EIO_WriteOnly | EIO_ReadWrite) ) != 0;
}

/** Test if the I/O Device is readable.
 * @return true if it is readable.
 */
inline bool EIODevice::isOpen() const
{
  return (info & EIO_IsOpen) != 0;
}

/** Test if the I/O Device is readable and writable.
  * @return true if it is readable and writable.
  */
inline bool EIODevice::isReadWrite() const
{
  return (info & EIO_ReadWrite) != 0;
}

/** Set access mode.
  * Set access mode for the device.
  * @param mode access mode.
  */
inline void EIODevice::setMode( int mode )
{
  EASSERT( (mode & ~EIO_ModeMask) == 0 );
  info = (info & ~EIO_ModeMask) | mode;
}

/** Set a status.
 * Set a status bit on.
 * @param state State to set.
 */
inline void EIODevice::setStatus( int state )
{
  info = (info & ~EIO_StatusMask) | state;
  info |= state;
}

/** Unset a status.
 * Unset a status bit on.
 * @param state State to unset.
 */
inline void EIODevice::unsetStatus( int state )
{
  info = (info & ~EIO_StatusMask) | state;
  info &= ~state;
}

/// reset access mode & status
inline void EIODevice::resetInfo()
{
    info = 0;
}

#define EIODEVIC_HPP_INCLUDED
#endif

// ########################### EOF EIODEVIC.HPP ##############################
// ########################### EOF EIODEVIC.HPP ##############################
// ########################### EOF EIODEVIC.HPP ##############################


