// ###########################################################################
// File: ELIST.HPP
// (c)Copyright 04/1997, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// Version 1.2
// ###########################################################################


#ifndef ELIST_HPP_INCLUDED


// Includes:
// #####################Last change: 07/04/97 11h04mn#########################


template<class type>
class EListROIterator;

// Constantes:
// #####################Last change: 07/04/97 11h04mn#########################


// Macros:
// #####################Last change: 07/04/97 11h04mn#########################
/** a "for" loop to walk through all element of a list.
    get the current element using the ptr() function
 */
#define EListForEach(itera) for ((itera).begin(); !(itera).noMore(); (itera).next())

// Types:
// #####################Last change: 07/04/97 11h04mn#########################

/// Double linked list (special optimization for at() method)
class EGenList {
  public:
    EGenList();
    ~EGenList();

    /// delete or remove all element from list (depending of the owner)
    void reset();

    /// insert an element at position pos (0 = beginning)
    void insertAt(void *elm, UL pos);
    /// insert an element at the end of the list
    void insertAtEnd(void *elm);
    /// insert an element at the beginning of the list
    void insertAtBeginning(void *elm);

    /** remove the element at position pos (0=first) in the list
        @return ptr on removed element
     */
    void *removeAt(UL pos);
    /** remove th first element of the list
     @return ptr on removed element
     */
    void *removeFirst();
    /** remove the last element of the list
     @return ptr on removed element
     */
    void *removeLast();

    /// delete the element at position pos (0=first) in the list
    void deleteAt(UL pos);
    /// delete the first element of the list
    void deleteFirst();
    /// delete the last element of the list
    void deleteLast();

    /** test if the list is empty
     @return TRUE if the list is empty
     */
    bool isEmpty();

    /** set owner of elementy
        owner -> TRUE if the list own the element, FALSE otherwise
     */
    void setOwner(bool owner);

    /** get element at position pos
     @return element at position pos, NULL if out of bound
     */
    void *at(UL pos);

    /** get element at position pos
     @return element at position pos, NULL if out of bound
     */
    void *operator [](UL pos);

    /** get number of element in the list
     @return number of element in the list
     */
    UL count() const;

#ifdef _SUPPORT_TEMPLATE_FRIEND
  protected:
#endif
    struct ENode {
      /// ptr on previous element of the list
      ENode *prev;
      /// ptr on next element of the list
      ENode *next;
      /// ptr on the element of the list
      void *elm;
    };

    /// ptr on head of list
    ENode *head;
    /// ptr on tail of list
    ENode *tail;
    /// ptr on current element
    ENode *cur;
    /// index of current element
    UL idxCur;
    /// number of element in the list
    UL nb;
    /// is the list owner of element ?
    bool mine;

    /// initialize list
    void init();

    /** give a ptr on the pos node of the list (0 is the first)
     @return ptr on node
     */
    ENode *nodeAt(UL pos);

    /** get the first node.
     * @return a pointer on the first node.
     */
    const ENode *firstNode() const;

    /** get the last node.
     * @return a pointer on the last node.
     */
    const ENode *lastNode() const;
};

/// typede Double linked list (interface to EGenList)
template<class type>
class EList : public EGenList {
#ifdef _SUPPORT_TEMPLATE_FRIEND
    friend class EListROIterator<type>;
#endif
  public:
    /// Alias for iterator type.
    typedef EListROIterator<type> EROIterator;

    /** remove the element at position pos (0=first) in the list
        @return ptr on removed element
     */
    type *removeAt(UL pos);

    /** remove th first element of the list
     @return ptr on removed element
     */
    type *removeFirst();
    /** remove the last element of the list
     @return ptr on removed element
     */
    type *removeLast();

    /** get element at position pos
     @return element at position pos, NULL if out of bound
     */
    type *at(UL pos);

    /** get element at position pos
     @return element at position pos, NULL if out of bound
     */
    type *operator [](UL pos);
};


/** this class is used to iterate through the list.
    The list can be modified but the current node
    MUST NOT be deleted.
 */
template<class type>
class EListROIterator {
  public:
    EListROIterator(const EList<type> *l);
    /// set iterator at the beginning of the list
    void begin();
    /// set iterator at the end of the list
    void end();
    /// return TRUE if there no more element
    bool noMore();
    /// seek to previous element
    void prev();
    /// seek to next element
    void next();
    /// return a ptr on the current element
    type *ptr();
  private:
    const EList<type> *liste;
    EGenList::ENode *cur;
};




// inlines:
// #####################Last change: 07/04/97 11h04mn#########################

// EGenList
inline bool EGenList::isEmpty() { return nb == 0; }
inline void EGenList::setOwner(bool owner) { mine = owner; }
inline void EGenList::insertAtEnd(void *elm) { insertAt(elm, nb); }
inline void EGenList::insertAtBeginning(void *elm) { insertAt(elm, 0); }
inline void *EGenList::operator [](UL pos) { return at(pos); }
inline void *EGenList::removeFirst() { return removeAt(0); }
inline UL EGenList::count() const { return nb; }
inline const EGenList::ENode *EGenList::firstNode() const { return head; }
inline const EGenList::ENode *EGenList::lastNode() const { return tail; }

// EList
template<class type>
inline type *EList<type>::removeAt(UL pos) {
  return (type *)EGenList::removeAt(pos);
}

template<class type>
inline type *EList<type>::removeFirst() {
  return (type *)EGenList::removeFirst();
}

template<class type>
inline type *EList<type>::removeLast() {
  return (type *)EGenList::removeLast();
}

template<class type>
inline type *EList<type>::at(UL pos) {
  return (type *)EGenList::at(pos);
}

template<class type>
inline type *EList<type>::operator [](UL pos) {
  return (type *)EGenList::operator [](pos);
}

// EListROIterator
template<class type>
inline EListROIterator<type>::EListROIterator(const EList<type> *l) : liste(l), cur(0) {
}

template<class type>
inline void EListROIterator<type>::begin() {
  cur = liste->firstNode();
}

template<class type>
inline void EListROIterator<type>::end() {
  cur = liste->lastNode();
}

template<class type>
inline void EListROIterator<type>::prev() {
  cur = cur ? cur->prev : 0;
}

template<class type>
inline void EListROIterator<type>::next() {
  cur = cur ? cur->next : 0;
}

template<class type>
inline type *EListROIterator<type>::ptr() {
  return (type *)(cur ? cur->elm : 0);
}

template<class type>
inline bool EListROIterator<type>::noMore() {
  return (cur == 0);
}

// Prototypes:
// #####################Last change: 07/04/97 11h04mn#########################



#define ELIST_HPP_INCLUDED
#endif


// ############################ EOF ELIST.HPP ################################
// ############################ EOF ELIST.HPP ################################
// ############################ EOF ELIST.HPP ################################

