// ###########################################################################
// File: EMATH.HPP
// (c)Copyright 10/1997, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// Version 1.2
// ###########################################################################


#ifndef EMATH_HPP_INCLUDED


// Includes:
// #####################Last change: 26/10/97 16h00mn#########################
#include <float.h>
#include <math.h>

// Constantes:
// #####################Last change: 26/10/97 16h00mn#########################

/* Constants rounded for 21 decimals. */
#define M_E         2.71828182845904523536
#define M_LOG2E     1.44269504088896340736
#define M_LOG10E    0.434294481903251827651
#define M_LN2       0.693147180559945309417
#define M_LN10      2.30258509299404568402
#define M_PI        3.14159265358979323846
#define M_2PI       (2.0*M_PI)
#define M_PI_2      1.57079632679489661923
#define M_PI_4      0.785398163397448309616
#define M_1_PI      0.318309886183790671538
#define M_2_PI      0.636619772367581343076
#define M_1_SQRTPI  0.564189583547756286948
#define M_2_SQRTPI  1.12837916709551257390
#define M_SQRT2     1.41421356237309504880
#define M_SQRT_2    0.707106781186547524401

// Size of sinus table.
// Size must be a power of 2.
const UL sinusTableSize = 1024;

// PI equivalent for sinus table.
const UL sinusPI = sinusTableSize/2;
const UL sinus2PI = sinusTableSize/4;
const UL sinusPI_2 = sinusTableSize/8;

// offset equivalent to PI/2
const UL cosinusOffset = sinusTableSize /4;

// mask equivalent to modulo 2*PI
const UL sinusMask = sinusTableSize-1;

// Unity
const SL unite = 1 << 16;


// Macros:
// #####################Last change: 26/10/97 16h00mn#########################

// Types:
// #####################Last change: 26/10/97 16h00mn#########################


// Globales:
// #####################Last change: 26/10/97 16h00mn#########################

// sinus table. -> lSinusTable[0] = unite...
extern SL lSinusTable[ sinusTableSize ];

// Prototypes:
// #####################Last change: 26/10/97 16h00mn#########################

/*! Initialize math stuff.
 */
void emathInit();

/// Initialize sinus table.
void lSinusInit();

/// initialize fast random number generator.
void fastRndInit( UL seed );

/// get a random number between 0 and 2^32. initFastRnd() must have been called.
UL fastRnd();

/// return a random number between 0.0 and 1.0
float frand();

/*! Return the absolute value of a number.
 * @param v Value.
 * @return Absolute value of \a v.
 */
template <class T>
T eabs( T x );

/*! Return the square of a value.
 * @param v Value.
 * @return Square of v: v*v.
 */
template <class T>
T esqr( T x );

/** Return the fractionnal part of a real number.
 * @param x Value
 * @return Fractionnal part of x. Example: efrac( 2.8 ) is 0.8.
 */
template <class T>
T efrac( T x );

/** Test if two double are equal use with a small tolerance.
 * @param flX First double.
 * @param flY Second double.
 * @return \c true if the absolute difference between \a flX and \a flY is
 *         less than a defined tolerance.
 */
bool eisEqual( double flX, 
               double flY );

/** Get the sign of a number.
 * @param fX Number we want the sign.
 * @return -1 if \a fX is less than 0, 1 if \a fX is greater than 0, and 
 *         0 if \a fX is 0.
 */
template <class T>
T esign( T fX );

// inlines:
// #####################Last change: 26/10/97 17h16mn#########################


/// return the value of the sinus for angle \a angle.
inline SL lSinus( UL angle )
{
  return lSinusTable[ angle & sinusMask ];
}


/// return value of the cosinus for angle \a angle.
inline SL lCosinus( UL angle )
{
  return lSinusTable[ (angle +cosinusOffset) & sinusMask ];
}


template <class T>
inline T esqr( T x )
{
  return x *x;
}


template <class T>
inline T eabs( T x )
{
  return x >= (T)0 ? x : -x;
}


template <class T>
inline 
T emax( T a, 
        T b )
{
  return a > b ? a : b;
}


template <class T>
inline 
T emin( T a, 
        T b )
{
  return a < b ? a : b;
}


template <class T>
inline 
T ebound( T v, 
          T minValue,
          T maxValue )
{
  return emin( emax( minValue, v ), maxValue );
}


template <class T>
inline T
efrac( T fX )
{
  T fAbsX = (T)fabs( fX );
  return fAbsX - (T)floor( fAbsX );
}

inline 
bool 
eisEqual( double flX, 
          double flY )
{
  double fDiff = fabs( flY - flX );
  return fDiff <= 1e-13;    // DBL_EPSILON;
}


template <class T>
inline T 
esign( T fX )
{
  if ( fX < (T)0 )
    return (T)-1;
  else if ( fX > (T)0 )
    return (T)1;
  return (T)0;
}

#define EMATH_HPP_INCLUDED
#endif

