// ###########################################################################
// File: EMIODEVI.CPP
// (c)Copyright 10/1997, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// version 0.3
// ###########################################################################


#include "usuel.hpp"
#include "emiodevi.hpp"


/// Default Constructor.
EMultiIODevice::EMultiIODevice() 
{
}

/// Destructor.
EMultiIODevice::~EMultiIODevice()
{
  if ( isOpen() )
    close();
  EIODevice::~EIODevice();
}

/** Add a device.
 * Only one instance of the device can be present in a multi I/O device.
 * If the device is already present, it will not be added.
 * @param device device to add.
 */
void EMultiIODevice::addDevice( EIODevice *device )
{
  assert( device );

  if ( !_aDevices.contains( &device ) )     // device is not already registred
    _aDevices.insert( _aDevices.count(), device );
}

/** Remove a device.
 * Remove a device.
 * @param device device to remove.
 */
void EMultiIODevice::removeDevice( EIODevice *device )
{
  int pos;
  if ( ( pos = _aDevices.find( &device ) ) >= 0 )
    _aDevices.remove( pos, 1 );
}

/** Return the number of devices.
 * @return number of devices.
 */
int EMultiIODevice::count() const
{
  return _aDevices.count();
}

/** Open the devices.
 * Open all the registred devices using mode \a _mode.
 * @param _mode mode use to open the device
 */
void EMultiIODevice::open( int _mode )
{
  SL cpt = _aDevices.count();

  while ( --cpt >= 0 )
  {
    _aDevices[cpt]->open( _mode );
  }
}

/** Close the devices.
 * Close all the registred devices.
 */
void EMultiIODevice::close()
{
  SL cpt = _aDevices.count();

  while ( --cpt >= 0 )
  {
    _aDevices[cpt]->close();
  }
}

/** Flush the devices.
 * Flush all the registred devices.
 */
void EMultiIODevice::flush()
{
}

/** Always return 0.
 * @return 0.
 */
SL EMultiIODevice::size() const
{
  return 0;
}

/** Always return 0.
 * @return 0
 */
SL EMultiIODevice::at() const
{
  return 0;
}

/** Set the position of devices.
 * Set the position of all the registred devices (index); to \a pos.
 * @param pos new position of the index.
 * @exception EXCIODevice on error.
 */
void EMultiIODevice::at( SL pos)
{
  SL cpt = _aDevices.count();

  while ( --cpt >= 0 )
  {
    _aDevices[cpt]->at( pos );
  }
}

/** always return false.
 * @return false.
 */
bool EMultiIODevice::atEnd() const
{
  return false;
}

/** Do nothing.
 * @param data dummy.
 * @param len dummy.
 * @return always 0.
 */
SL EMultiIODevice::readBlock( void *data, SL len)
{
  return 0;
}

/** Write a block of byte on devices.
 * Write \a len Byte of buffer \a data on each registred devices.
 * @param data buffer where data are read.
 * @param len number of bytes of data to write.
 * @exception EXCIODevice is case of grave error.
 */
void EMultiIODevice::writeBlock( const void *data, SL len )
{
  SL cpt = _aDevices.count();

  while ( --cpt >= 0 )
  {
    _aDevices[cpt]->writeBlock( data, len );
  }
}

/** Always return 0.
 * @return 0.
 */
UB EMultiIODevice::getByte()
{
  return 0;
}

/** Write a Byte on each devices.
 * @param c byte to write.
 * @exception EIODevice on error.
 */
void EMultiIODevice::putByte( UB c )
{
  SL cpt = _aDevices.count();

  while ( --cpt >= 0 )
  {
    _aDevices[cpt]->putByte( c );
  }
}

/** Push-back a byte on each devices.
 * @param c byte to push-back.
 * @exception EIODevice on error.
 */
void EMultiIODevice::ungetByte( UB c )
{
  SL cpt = _aDevices.count();

  while ( --cpt >= 0 )
  {
    _aDevices[cpt]->ungetByte( c );
  }
}

// ########################### EOF EMIODEVI.CPP ##############################
// ########################### EOF EMIODEVI.CPP ##############################
// ########################### EOF EMIODEVI.CPP ##############################


