// ###########################################################################
// File: ESTREAM.CPP
// (c)Copyright 04/1998, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// ###########################################################################

#include "usuel.hpp"
#include "estream.hpp"

/* HST:
 * 04/04/98: - default byte order is now littleEndian.
 *           - bug fix: when printable, no return in readBLock & writeBlock.
 *           - bug fix: writeBlock() when printable, read instead of write.
 *           - bug fix: readInteger(). Does not do \0 end of tmp.
 *           - add putByte() & getByte().
 * 05/04/98: - bug fix: _bIsPrintable was not initialized in constructor.
 */

/* TODO:
 * - revoir flush(), implementation plus generique.
 */

/* Test:
 * [  ] EStream()
 * [OK] EStream( EIODevice *pDevice )
 * [OK] ~EStream()
 * [OK] void setDevice( EIODevice *pDevice )
 * [OK] EIODevice *device() const
 * [  ] void unsetDevice()
 * [OK] void setByteOrder( ByteOrder byteOrder )
 * [OK] EStream::ByteOrder byteOrder() const
 * [  ] void setPrintableStream( bool bPrintableStream =true )
 * [  ] bool isPrintable() const
 * [OK] bool needSwap() const
 * [  ] void writeInteger( SL v )
 * [  ] SL readInteger()
 * [  ] void writeDouble( double v )
 * [  ] double readDouble()
 * [  ] EStream &getPrintableByte( SB& v )
 * [  ] EStream &putPrintableByte( SB v )
 * [OK] void init()
 */

/// initialisation
void EStream::init()
{
  _pDevice =ENIL;
  setByteOrder( littleEndian );
#ifdef PRINTABLE_STREAM
  _bIsPrintable = false;
#endif
}

/// Constructs a data stream with no device.
EStream::EStream()
{
  init();
}

/** Constructs a data stream with a device.
  * Constructs a data stream for device \a device.
  * @param device device used by the data stream.
  */
EStream::EStream( EIODevice *pDevice )
{
  init();
  setDevice( pDevice );   // note: call EStream::setDevice, even if overloaded.
}

/// Destructor.
EStream::~EStream()
{
  unsetDevice();
}

/** Set device used by the data stream.
  * Set the device used by the data stream.
  * @param pDevice device used by the data stream.
  */
void EStream::setDevice( EIODevice *pDevice )
{
  _pDevice = pDevice;
}

/** Unset device used by the data stream.
  * Unset the device used by the data stream.
  */
void EStream::unsetDevice()
{
  if ( device() )
  {
    flush();
    _pDevice =ENIL;
  }
}

/** Set the byte order.
  * Set the byte order to \a byteorder.
  * @param byteOrder byte order used for data storage.
  */
void EStream::setByteOrder( ByteOrder byteOrder )
{
                                // determine byte order
  UW v16 = 0x1234;
  ByteOrder thisOrder;

  if ( *(UB *) &v16 == 0x12 )
    thisOrder = bigEndian;
  else
    thisOrder = littleEndian;

  _bNeedSwap = ( byteOrder != thisOrder );
}

/// write an integer as text
void EStream::writeInteger( SL v )
{
  char tmp[ 40 ];
  sprintf( tmp, "%d", v );
  device()->writeBlock( tmp, strlen( tmp ) );
}

/// write an integer as text
SL EStream::readInteger()
{
  char tmp[40];
  int p;
  for ( p=0; ; ++p)
  {
    assert( p < 40 );
    tmp[ p ] = device()->getByte();
    if ( tmp[ p ] == '\n' )
    {
      tmp[p] = 0;
      break;
    }
  }
  return atol( tmp );
}

/// write a double as text
void EStream::writeDouble( double v )
{
  char tmp[ 80 ];
  sprintf( tmp, "%g", v );
  device()->writeBlock( tmp, strlen( tmp ) );
}

/// read a double from text
double EStream::readDouble()
{
  char tmp[80];
  int p;
  for ( p=0; ; ++p)
  {
    assert( p < 80 );
    tmp[ p ] = device()->getByte();
    if ( tmp[ p ] == '\n' )
      break;
  }
  return atof( tmp );
}

#ifdef PRINTABLE_STREAM
/** Get a printable byte from the stream.
 * @param on return, byte read.
 * @return reference on this stream.
 */
EStream &EStream::getPrintableByte( SB& v )
{
  v = device()->getByte();
  if ( v == '\\' )          // read octal code
  {
    char tmp[3];
    device()->readBlock( tmp, sizeof( tmp ) );
    v = (tmp[2] & 7) + ((tmp[1] & 7) << 3) +((tmp[0] & 7) << 6);
  }
  return *this;
}

/** Put a printable byte in the stream.
 * @param v byte to put.
 * @return reference on this stream.
 */
EStream &EStream::putPrintableByte( SB v )
{
  if ( isPrintable() ||  v == '\\' )  // print octal code
  {
    char tmp[4];
    tmp[0] = '\\';
    tmp[1] = '0' + ((v>>6) & 7);
    tmp[2] = '0' + ((v>>3) & 7);
    tmp[3] = '0' + (v & 7);
    device()->writeBlock( tmp, sizeof( tmp ) );
  }
  else
    device()->putByte( v );
  return *this;
}
#endif



// ########################### EOF ESTREAM.CPP ###############################
// ########################### EOF ESTREAM.CPP ###############################
// ########################### EOF ESTREAM.CPP ###############################

