// ###########################################################################
// File: ETEXTSTR.CPP
// (c)Copyright 10/1997, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// version 0.3
// ###########################################################################


/* History:
 * 27/01/98: Bug dans writeInt, - mode hex (&1 au lieu de &15).
 *                              - tous les modes: shift gauche au lieu droite.
 * 03/06/2000: - BugFix: manipulator now works (endl...)
 *             - Added: writeLine(), readLine(), getEol().
 *             - BugFix: writeBlock() works if the block is NULL.
 */

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include "usuel.hpp"
#include "etextstr.hpp"
#include "EString.hpp"

/** TODO:
 * - write/read double
 * - good managment of end of file.
 * - set _strEol to OS default...
 */



/// initialisation
void ETextStream::init()
{
  mDev =0;
  _strEol = "\r\n";
  reset();
}

/// Constructs a data stream with no device.
ETextStream::ETextStream()
{
  init();
}

/** Constructs a data stream with a device.
  * Constructs a data stream for device \a device.
  * @param device device used by the data stream.
  */
ETextStream::ETextStream( EIODevice *device )
{
  init();
  setDevice( device );
}

/// Destructor.
ETextStream::~ETextStream()
{
}

/** Set device used by the data stream.
  * Set the device used by the data stream.
  * @param device device used by the data stream.
  */
void ETextStream::setDevice( EIODevice *device )
{
  mDev = device;
}

/** Unset device used by the data stream.
  * Unset the device used by the data stream.
  */
void ETextStream::unsetDevice()
{
  mDev =0;
}

/** Read a block.
  * @param data data to read.
  * @param len length of data to read in byte.
  */
void ETextStream::readBlock( void *data, SL len )
{
  assert( len >= 0 );
  mDev->readBlock( data, len );
}

/** Write a block.
  * @param data data to write.
  * @param len length of data to write in byte.
  */
void ETextStream::writeBlock( const void *data, SL len )
{
  assert( len >= 0 );
  mDev->writeBlock( data, len );
}


/** Write an integer as text.
 * Write integer \a v in text form. if \pisNeg is true then
 * a negative sign is added if decimal.
 * \a size is the length of the integer in byte (for binary format).
 * @param v value
 * @param isNeg is the integer negativ.
 * @param size length of the integer in byte.
 */
void ETextStream::writeInt( UL v, bool isNeg, UL size )
{
  char buf[80];
  char *p = &buf[80];
  static char hexTable[17] = "0123456789ABCDEF";

  *--p = 0;

  switch ( flags() & basef )
  {
    case dec:
        if ( isNeg )
          v = (UL)( -(SL)v );  // change to positive

        do
        {
          *--p = (char)( (v % 10) +'0' );
          v /= 10;
        }
        while ( v );

        if ( isNeg )
          *--p = '-';
        break;
    case bin:
        do
        {
          *--p = (char)( (v & 1) +'0' );
          v >>= 1;
        }
        while ( v );
        *--p = 'b';
        *--p = '0';
        break;
    case oct:
        do
        {
          *--p = (char)( (v & 7) +'0' );
          v >>= 3;
        }
        while ( v );
        *--p = '0';
        break;
    case hex:
        do
        {
          *--p = hexTable[ v & 15 ];
          v >>= 4;
        }
        while ( v );
        *--p = 'x';
        *--p = '0';
        break;
  }

  *this << p;
}

/// read a decimal from text
UL ETextStream::readDec()
{
  UL v = 0;
  char c;

  while ( isdigit( c = mDev->getByte() ) )
  {
    v *= 10;
    v += c -'0';
  }
  return v;
}

/// read a hexadecimal from text
UL ETextStream::readHex()
{
  UL v = 0;
  char c;

  while ( isxdigit( c = mDev->getByte() ) )
  {
    v >>= 4;
    v += tolower(c) - ( isdigit( c ) ? '0' : 'a' );
  }
  return v;
}

/// read an octal from text
UL ETextStream::readOct()
{
  UL v = 0;
  char c;

  while ( ( c = mDev->getByte() )  &&  c >= '0'  &&  c <='7' )
  {
    v >>= 3;
    v += c - '0';
  }
  return v;
}

/// read a binary from text
UL ETextStream::readBin()
{
  UL v = 0;
  char c;

  while ( ( c = mDev->getByte() )  &&  c >= '0'  &&  c <='1' )
  {
    v >>= 1;
    v += c - '0';
  }
  return v;
}

UL ETextStream::readInt()
{
  UL v;
  char c;

  switch ( flags() & basef )
  {
    case hex:
        return readHex();
    case oct:
        return readOct();
    case bin:
        return readBin();
    case dec:
        c = mDev->getByte();
        if ( !( c == '-'  ||  c == '+' ) )
        {
          mDev->ungetByte( c );
          return readDec();
        }
        else
        {
          v = readDec();
          if ( c == '-' )
            return -(SL)v;
        }

    default:   // guess format using 0x 0b...
        return 0;
  }
}

/// write a double as text
void ETextStream::writeDouble( double v )
{
  char buf[80];
  sprintf( buf, "%g", v );
  *this << buf;
}

/// read a double from text
double ETextStream::readDouble()
{
  return 0.0;
}

ETextStream& ETextStream::operator >>( char &v )
{
  if ( flags() & skip )
    v = skipSpaces();
  else
    v = (char) mDev->getByte();
  return *this;
}

ETextStream& ETextStream::operator >>( SW &v )
{
  v = (SW) readInt();
  return *this;
}

ETextStream& ETextStream::operator >>( SL &v )
{
  v = (SL) readInt();
  return *this;
}

ETextStream& ETextStream::operator >>( UW &v )
{
  v = (UW) readInt();
  return *this;
}

ETextStream& ETextStream::operator >>( UL &v )
{
  v = (UL) readInt();
  return *this;
}

ETextStream& ETextStream::operator >>( float &v )
{
  v = (float) readDouble();
  return *this;
}

ETextStream& ETextStream::operator >>( double &v )
{
  v = readDouble();
  return *this;
}

/// read a word. v must be preallocated.
ETextStream& ETextStream::operator >>( char *v )
{
  char c;
  SL maxLen = width();

  if ( !maxLen )
    maxLen = -1;

  if ( flags() & skip )
    c = skipSpaces();
  else
    c = (char) mDev->getByte();

  while ( isspace(c)  || maxLen-- )
  {
    *v++ = c;
    c = mDev->getByte();
  }
  mDev->ungetByte( c );
  *v++ =0;
  return *this;
}

ETextStream& ETextStream::operator <<( char v )
{
  mDev->putByte( v );
  return *this;
}

ETextStream& ETextStream::operator <<( SW v )
{
  writeInt( v, v < 0, 16 );
  return *this;
}

ETextStream& ETextStream::operator <<( SL v )
{
  writeInt( v, v < 0, 32 );
  return *this;
}

ETextStream& ETextStream::operator <<( UW v )
{
  writeInt( v, false, 16 );
  return *this;
}

ETextStream& ETextStream::operator <<( UL v )
{
  writeInt( v, false, 32 );
  return *this;
}

ETextStream& ETextStream::operator <<( float v )
{
  writeDouble( v );
  return *this;
}

ETextStream& ETextStream::operator <<( double v )
{
  writeDouble( v );
  return *this;
}

ETextStream& ETextStream::operator <<( const char *v )
{
  if ( !v )
    return *this;

  SL len = strlen( v );

  if ( mWidth )   // do padding
  {
    SL padLen = mWidth -len;
    if ( padLen > 0 )        // need padding
    {
      if ( mFlag & left )
      {
        writeBlock( v, len );
        while ( padLen-- )
          *this << mFiller;
      }
      else
      {
        while ( padLen-- )
          *this << mFiller;
        writeBlock( v, len );
      }
      return *this;
    }
  }

  writeBlock( v, len );  // save string
  return *this;
}


/// reset TextStream to default settings
void ETextStream::reset()
{
  mFlag = 0;
  setFlag( dec, basef );
  setFlag( skip );
  setFlag( left, padf );
  fill( ' ' );
  width( 0 );
}

/// skip white space and return char.
char ETextStream::skipSpaces()
{
  char c;

  while ( isspace( c = mDev->getByte() ) );

  return c;
}


void
ETextStream::readLine( EString &strLine )
{
  strLine.setNull();
  bool bReadSomething = false;
  ETRY 
  {
    char cAhead =0;
    bool bOneCharAhead =false;
    do
    {
      char cChar = bOneCharAhead ? cAhead : 
                                   mDev->getByte();
      bOneCharAhead = false;
      bReadSomething = true;

      // Convert Dos & Mac New Line to Unix '\n'
      // New Line:
      // 13,10 : dos (\r\n)
      // 10 : unix (\n)
      // 13 : mac (\r)
      if ( cChar == '\r' )        // Dos or Mac   : convert new line sequence to 1 \n
      {
        cAhead = mDev->getByte();
        if ( cAhead != '\n' )     // Mac
          bOneCharAhead = true;
        cChar = '\n';
      }

      // End of line ?
      if ( cChar == '\n' )   // Unix, Dos, Mac New Line
        break;

      // Append char to line
      strLine += cChar;
    }
    while ( true );

    if ( bOneCharAhead )
      mDev->ungetByte( cAhead );
  }
  ECATCH( EXCIODevice, e )
  {
    if ( e.cause() != EXCIODevice::endOfFile  ||
         !bReadSomething )
      ETHROW_AGAIN;
  }
}


void 
ETextStream::writeLine( const char *cszLine )
{
  *this  <<  cszLine  <<  endl;
}
