// ###########################################################################
// File: ETEXTSTR.HPP
// (c)Copyright 10/1997, Lepilleur Baptiste (Aka Elyon from Zuul Design)
// version 0.3
// ###########################################################################



#ifndef ETEXTSTR_HPP_INCLUDED


// Includes:
// #####################Last change: 15/10/97 15h52mn#########################

#include "eiodevic.hpp"
#include "EString.hpp"

// Types:
// #####################Last change: 15/10/97 15h52mn#########################

class ETextStream
{
  public:
    /// Constructs a data stream with no device.
    ETextStream();

    /** Constructs a data stream with a device.
     * Constructs a data stream for device \a device.
     * @param device device used by the data stream.
     */
    ETextStream( EIODevice *device );

    /// Destructor.
    ~ETextStream();

    /** Set device used by the data stream.
     * Set the device used by the data stream.
     * @param device device used by the data stream.
     */
    void setDevice( EIODevice *device );

    /** Return the device.
     * @return device used by the data stream.
     */
    EIODevice *device() const;

    /** Unset device used by the data stream.
     * Unset the device used by the data stream.
     */
    void unsetDevice();

    /** Read a block.
     * @param data data to read.
     * @param len length of data to read in byte.
     */
    void readBlock( void *data, SL len );

    /** Write a block.
     * @param data data to write.
     * @param len length of data to write in byte.
     */
    void writeBlock( const void *data, SL len );

    ETextStream& operator >>( char &v );
    ETextStream& operator >>( SW &v );
    ETextStream& operator >>( SL &v );
    ETextStream& operator >>( UW &v );
    ETextStream& operator >>( UL &v );
    ETextStream& operator >>( float &v );
    ETextStream& operator >>( double &v );
    ETextStream& operator >>( char * v );
    ETextStream& operator >>( ETextStream &(*manipultor)(ETextStream &) );

    ETextStream& operator <<( char v );
    ETextStream& operator <<( SW v );
    ETextStream& operator <<( SL v );
    ETextStream& operator <<( UW v );
    ETextStream& operator <<( UL v );
    ETextStream& operator <<( float v );
    ETextStream& operator <<( double v );
    ETextStream& operator <<( const char *v );
    ETextStream& operator <<( ETextStream &(*manipultor)(ETextStream &) );

    enum {
      bin  = 0x0001,
      oct  = 0x0002,
      dec  = 0x0004,
      hex  = 0x0008,
      basef= (bin|oct|dec|hex),
      skip = 0x0010,
      left = 0x0020,
      right= 0x0040,
      padf = (left|right)
    };

    /// return filling char.
    char fill() const;

    /// set filling char and return old.
    char fill( char filler );

    /// return field width.
    SL width() const;

    /// set field width for padding and return old.
    SL width( SL w );


    void setFlag( SL flag, SL mask );
    void setFlag( SL flag );
    void putFlag( SL Flag );
    void unsetFlag( SL flag );
    SL flags() const;


    /// reset TextStream to default settings
    void reset();

    /// skip white space.
    void ws();

    void readLine( EString &strLine );

    void writeLine( const char *cszLine =NULL );

    /*!
     * Get the end of line sequence.
     */
    const EString &getEol() const;

  private:
    /// initialisation
    void init();


    /** Write an integer as text.
     * Write integer \a v in text form. if \pisNeg is true then
     * a negative sign is added if decimal.
     * \a size is the length of the integer in byte (for binary format).
     * @param v value
     * @param isNeg is the integer negativ.
     * @param size length of the integer in byte.
     */
    void writeInt( UL v, bool isNeg, UL size );

    /// read an integer from text
    UL readDec();
    /// read a hexadecimal from text
    UL readHex();
    /// read an octal from text
    UL readOct();
    /// read a binary from text
    UL readBin();
    /// read a double from text
    double readDouble();
    UL readInt();

    /// write a double as text
    void writeDouble( double v );


    /** skip white space and return char.
     * @exception EXCIODevice -> end of file
     */
    char skipSpaces();

    /// device used by the data stream.
    EIODevice *mDev;

    /// base & skip space.
    SL mFlag;

    /// field width
    SL mWidth;

    /// character used for filling.
    char mFiller;

    /*!
     * End of line sequence.
     */
    EString _strEol;
};


// Globales:
// #####################Last change: 15/10/97 15h52mn#########################

ETextStream& dec( ETextStream& s );     // output integer in base 10
ETextStream& hex( ETextStream& s );     // output integer in base 16
ETextStream& oct( ETextStream& s );     // output integer in base 8
ETextStream& bin( ETextStream& s );     // output integer in base 2
ETextStream& reset( ETextStream& s );   // reset to default settings
ETextStream& endl( ETextStream& s );    // insert \n
ETextStream& flush( ETextStream& s );   // flush device
ETextStream& ws( ETextStream& s );      // skip white space

// inlines:
// #####################Last change: 15/10/97 16h45mn#########################

/** Return the device.
  * @return device used by the data stream.
  */
inline EIODevice *ETextStream::device() const
{
  return mDev;
}

inline SL ETextStream::flags() const
{
  return mFlag;
}

inline void ETextStream::putFlag( SL flag )
{
  mFlag = flag;
}

inline void ETextStream::setFlag( SL flag, SL mask )
{
  mFlag = (mFlag & ~mask) | flag;
}

inline void ETextStream::setFlag( SL flag )
{
  mFlag |= flag;
}

inline void ETextStream::unsetFlag( SL flag )
{
  mFlag &= ~flag;
}

inline ETextStream& dec( ETextStream& s )
{
  s.setFlag( ETextStream::dec, ETextStream::basef );
  return s;
}

inline ETextStream& hex( ETextStream& s )
{
  s.setFlag( ETextStream::hex, ETextStream::basef );
  return s;
}

inline ETextStream& oct( ETextStream& s )
{
  s.setFlag( ETextStream::oct, ETextStream::basef );
  return s;
}

inline ETextStream& bin( ETextStream& s )
{
  s.setFlag( ETextStream::bin, ETextStream::basef );
  return s;
}

inline ETextStream& reset( ETextStream& s )
{
  s.reset();
  return s;
}

inline ETextStream& endl( ETextStream& s )
{
  s << s.getEol();
  return s;
}

inline ETextStream& flush( ETextStream& s )
{
  s.device()->flush();
  return s;
}

inline ETextStream& ws( ETextStream& s )
{
  s.setFlag( ETextStream::skip );
  return s;
}

inline char ETextStream::fill() const
{
  return mFiller;
}

inline char ETextStream::fill( char filler )
{
  char oldF = mFiller;
  mFiller = filler;
  return oldF;
}

/// return field width.
inline SL ETextStream::width() const
{
  return mWidth;
}

/// set field width for padding.
inline SL ETextStream::width( SL w )
{
  SL oldW = mWidth;
  mWidth = w;
  return oldW;
}

/// skip white space.
inline void ETextStream::ws()
{
  try
  {
    mDev->ungetByte( skipSpaces() );
  }
  catch ( EXCIODevice * /*e*/)
  {
  }
}


inline const EString &
ETextStream::getEol() const
{
  return _strEol;
}


inline ETextStream& 
ETextStream::operator <<( ETextStream &(*manipulator)(ETextStream &) )
{
  return (*manipulator)( *this );
}


inline ETextStream& 
ETextStream::operator >>( ETextStream &(*manipulator)(ETextStream &) )
{
  return (*manipulator)( *this );
}



#define ETEXTSTR_HPP_INCLUDED
#endif
