// TfwkClassId.h: interface for the TfwkClassId class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TFWKCLASSID_H__43290692_4193_11D3_BE29_0040053AD648__INCLUDED_)
#define AFX_TFWKCLASSID_H__43290692_4193_11D3_BE29_0040053AD648__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000


#include "TfwkCommon.h"


class TfwkClass;

/*!
 * This macro provides the class id object for a given class.
 * @param ClassName Name of the class for which the class id is needed.
 * \sa TfwkClassId, TFWKDECL_DYNCLASS, TFWKIMPL_DYNCLASS.
 */
#define TFWKCLASSID( ClassName ) ClassName::_clsid


/*!
 * This macro is use to declare that a class support dynamic creation & identification.
 * \sa TfwkClassId, TFWKIMPL_DYNCLASS, TFWKCLASSID.
 */
#define TFWKDECL_DYNCLASS                                              \
  static const TfwkClassId _clsid;                                     \
  static TfwkClass *createObject();                                    \
  static const TfwkClassId &clsid() { return _clsid; }                 \
  virtual const TfwkClassId &objectId() const { return clsid(); }
    
/*!
 * This macro is use to implement dynamic creation & identification of a class.
 *
 * If a class has been declared to support dynamic creation, this macro must
 * be use in the implementation file.
 * \sa TfwkClassId, TFWKDECL_DYNCLASS, TFWKCLASSID.
 */
#define TFWKIMPL_DYNCLASS( Class, ParentClass )                             \
  TfwkClass *Class::createObject()                                          \
  {                                                                         \
    return new Class;                                                       \
  }                                                                         \
  const TfwkClassId TFWKCLASSID(Class)( #Class,                             \
                                        #ParentClass,                       \
                                        Class::createObject ) 

/*! \class TfwkClassId
 * \include "TfwkClassId.h"
 * \brief This class provides class identification capability.
 *
 * This class provides identification and dynamic creation functionnality.
 *
 * The following example shows how to use this:
 *
 * \code
 * // The header file (MyClass.h)
 * 
 * class MyClass : public TfwkClass
 * {
 * public:
 *   TFWKDECL_DYNCLASS;   // macro uses to declare support for dynamic creation.
 *   MyClass() {}         // dynamic creation only require a default constructor.
 * };
 *
 *
 * // The implementation file (MyClass.cpp)
 *
 * TFWKIMPL_DYNCLASS( MyClass, TfwkClass );  // macro to implement dynamic creation
 *
 * // Somewhere...
 * { // get the class id object
 *   const TfwkClassId &clsidMyClass = TFWKCLASSID( MyClass );  
 *
 *   // create an object using the class id.
 *   MyClass *pObject = (MyClass *)clsidMyClass.createObject();
 *
 *   // identify an object class.
 *   pObject->objectId().nameClass();
 *
 *   // identify an object parent class.
 *   pObject->objectId().nameParentClass();
 * }
 * \endcode
 * \sa TFWKDECL_DYNCLASS, TFWKIMPL_DYNCLASS, TFWKCLASSID.
 */
class TFWK_UI_EXPORT TfwkClassId  
{
public:
  /*!
   * Constructs a new class id.
   * @param szNameClass Name of the class this class id is defined for. 
   *                    Must not be \c NULL.
   * @param szNameParentClass Name of the parent class of this class. 
   *                          May be \c NULL.
   * @param fnCreateObject Function used to create a new object of this class.
   *                       May be \c NULL.
   */
	TfwkClassId( const char *szNameClass,
               const char *szNameParentClass,
               TfwkClass *(*fnCreateObject)() );

  /*!
   * Destructor.
   */
  ~TfwkClassId() {}

  /*!
   * Gets a pointer on the first class id of the linked list.
   * @return Pointer on the first class id of the linked list. May be \c NULL.
   */
  static TfwkClassId *firstClassId();

  /*!
   * Gets a pointer on the next class id of the linked list.
   * @return Pointer on the next class id of the linked list.
   */
  TfwkClassId *nextClassId() const;

  /*!
   * Gets the name of the class.
   * @return Name of the class. Never \c NULL.
   */
  const char *nameClass() const;

  /*!
   * Gets the name of the parent class.
   * @return Name of the parent class. Never \c NULL.
   */
  const char *nameParentClass() const;

  /*!
   * Creates an object of this type.
   * @return Pointer on the created object.
   */
  TfwkClass *createObject() const;

  /*! Finds the class id for a given class name.
   * @param szNameClass Class name of the class id that is searched.
   */
  static TfwkClassId *findClassId( const char *szNameClass );

  /*! Finds the class id of all class which are derived from a given class.
   * @param szNameParentClass Name of the parent class. May be \c NULL.
   * @param pCurClassId Pointer on the last class id that matched. The
   *                    search will be begin with the next class id.
   *                    To start a search, set \a pCurClassId to \c NULL.
   * @return Pointer on the first class id following \a pCurClassId for which
   *         the name of the parent class matched. \c NULL if no match found.
   */
  static TfwkClassId *findInheritedClassId( const char *szNameParentClass,
                                            TfwkClassId *pClassIdPrev =NULL );


private:
  /*! Linked list of all class id. 
   * (Not used anymore because could be made thread specific (TLS))
   */
//  static TfwkClassId *_pClassIdFirst;

  /*! Next class id in the linked list.
   */
  TfwkClassId *_pClassIdNext;

  /*! Class name.
   */
  const char *_szNameClass;

  /*! Parent class name.
   */
  const char *_szNameParentClass;

  /*! Object creator for this class.
   */
  TfwkClass *(*_fnCreateObject)();

};


// Inline : TfwkClassId
// --------------------

inline 
TfwkClassId::TfwkClassId( const char *szNameClass,
                          const char *szNameParentClass,
                          TfwkClass *(*fnCreateObject)() ) :
    _szNameClass( szNameClass ),
    _szNameParentClass( szNameParentClass ),
    _fnCreateObject( fnCreateObject )
{
}


inline TfwkClassId *
TfwkClassId::nextClassId() const
{
  return _pClassIdNext;
}


inline const char *
TfwkClassId::nameClass() const
{
  return _szNameClass ? _szNameClass : "";
}


inline const char *
TfwkClassId::nameParentClass() const
{
  return _szNameParentClass ? _szNameParentClass : "";
}


inline TfwkClass *
TfwkClassId::createObject() const
{
  TFWKASSERT( _fnCreateObject );
  return _fnCreateObject();
}


#endif // !defined(AFX_TFWKCLASSID_H__43290692_4193_11D3_BE29_0040053AD648__INCLUDED_)
