// //////////////////////////////////////////////////////////////////////////
// Implementation file TtrFmtTextBuilder.cpp for class TtrFmtTextBuilder
// (c)Copyright 2000, Baptiste Lepilleur.
// Created: 2000/09/16
// //////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "TtrFmtTextBuilder.h"
#include "TtrTextLine.h"
#include "TtrFmtText.h"
#include "TtrFmtBlank.h"
#include "TtrFmtPara.h"
#include "TtrWriterSetting.h"
#include "TtrFmtEltAdapter.h"


// //////////////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////////////
// class TtrFmtTextBuilder::VisitorIndent
// //////////////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////////////
class TtrFmtTextBuilder::VisitorIndent : public TtrFmtEltVisitor
{
public:
  VisitorIndent( const TtrTextLine &line ) : _line( line ) {}

  virtual void visitBlank( TtrFmtBlank &blank )
  {
    TtrFmtIndentInfo info;
    makeIndentInfo( info );
    blank.setIndent( info );
  }

  virtual void visitPara( TtrFmtPara &para )
  {
    TtrFmtIndentInfo info;
    makeIndentInfo( info );
    if ( info.indentLevel() < 0 )   // case first line is glued !!!
      info.setIndentLevel( 0 );   
    para.setIndent( info );
  }

private:
  void makeIndentInfo( TtrFmtIndentInfo &info )
  {
    info.setIndentLevel( _line.indentLevel() );
    info.setIndentString( _line.left( _line.indentLength() ) );
  }

private:
  const TtrTextLine &_line;
};


// //////////////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////////////
// class TtrFmtTextBuilder::VisitorAddLine
// //////////////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////////////
class TtrFmtTextBuilder::VisitorAddLine : public TtrFmtEltVisitor
{
public:
  VisitorAddLine( const TtrTextLine &line,
                  bool bRemoveEndingHyphen ) : 
                         _line( line ), 
                         _bRemoveEndingHyphen( bRemoveEndingHyphen ) {}

  virtual void visitBlank( TtrFmtBlank &blank )
  {
    blank.addLine( _line.tagList() );
  }

  virtual void visitPara( TtrFmtPara &para )
  {
    EString strContent( _line.mid( _line.indentLength(), -1 ) );
    strContent.trimSpaces();
    if ( para.content().length() )    // adds a space or 
    {                                 // concat if ending with '-'
      EString strPara = para.content();
      if ( _bRemoveEndingHyphen  &&  strPara.lastChar() == '-' )
        strPara.truncate( strPara.length() -1 );
      else
        strPara += ' ';
      strPara += strContent;
      para.setContent( strPara );
    }
    else
      para.setContent( strContent );
  }

private:
  const TtrTextLine &_line;
  bool _bRemoveEndingHyphen;
};




// //////////////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////////////
// class TtrFmtTextBuilder
// //////////////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////////////


TtrFmtTextBuilder::TtrFmtTextBuilder( TtrFmtText &text,
                                      const TtrWriterSetting &setting ) :
    _pText( &text ),
    _bRemoveEndingHyphen( setting.getWriterRemoveEndingHyphen() ),
    _bJumpOneLine( setting.getWriterJumpOneLine() )
{
}


TtrFmtTextBuilder::~TtrFmtTextBuilder()
{
}



void 
TtrFmtTextBuilder::buildBlank( const TtrTextLine &line )
{
  // Reuses current "blank" block or creates a new one if none exists.
  if ( _pText->eltCount() > 0 )
  {
    _ptrElt = _pText->takeEltAt( _pText->eltCount() -1 );
    if ( _ptrElt->type() != TtrFmtElt::eltBlank )
       _pText->addElt( _ptrElt );
  }

  if ( _ptrElt == NULL )
    _ptrElt = TtrFmtEltPtr( new TtrFmtBlank() );

  // Adds line to blank block...
  TtrFmtBlank &blank = *(TtrFmtBlank *)_ptrElt.get();
  if ( blank.lineCount() == 0)
    setIndent( line );
  
  blank.addLine( line.tagList() );
}


void 
TtrFmtTextBuilder::buildPara( const TtrTextLine &line )
{
  EAutoPtrT<TtrFmtPara> ptrPara( new TtrFmtPara() );
  ptrPara->tagList().copy( line.tagList() );
  _ptrElt = TtrFmtEltPtr( ptrPara.detach() );
  setIndent( line );
  addLineToElt( line );
}


void 
TtrFmtTextBuilder::setIndent( const TtrTextLine &line )
{
  getElt().visit( VisitorIndent( line ) );
}


void 
TtrFmtTextBuilder::addLineToElt( const TtrTextLine &line )
{
  getElt().visit( VisitorAddLine( line, _bRemoveEndingHyphen ) );
}


TtrFmtElt &
TtrFmtTextBuilder::getElt()
{
  EASSERT( (bool)_ptrElt );
  return *_ptrElt.get();
}


void 
TtrFmtTextBuilder::addElt()
{
  bool bJumpLine = _bJumpOneLine  && 
                   ( _ptrElt != NULL  &&  
                     _ptrElt->type() == TtrFmtElt::eltPara );


  _pText->addElt( _ptrElt );

  if ( bJumpLine )
  {
    TtrFmtPara *pPara = (TtrFmtPara *)(&_pText->eltAt( _pText->eltCount() -1 ) );

    // Adds a blank line wiht the same indentation of the added paragraph.
    EAutoPtrT<TtrFmtBlank> ptrBlank( new TtrFmtBlank() );
    ptrBlank->setIndent( pPara->indent() );
    ptrBlank->addLine( TtrTagList::nullObject );
    _ptrElt = TtrFmtEltPtr( ptrBlank.detach() );
    _pText->addElt( _ptrElt );
  }
}


void 
TtrFmtTextBuilder::dismissElt()
{
  _ptrElt.reset();
}

