/* RuleContext
 * Weirds thing may happen if current line is 0...
 * Should implements a TtrTextLineNull using the NullObject pattern.
 * Should also study if it may be possible to use this object as a caching proxy
 * for the text. Caching would included queries such as line.isEos()... which
 * are time consumming. On another hand, this object is rather volatile and the
 * cache would not be able to carry on from one reformatting to another one.
 */

#include "StdAfx.h"
#include "TtrTextFormaterV0.h"
#include "TtrGlobalSetting.h"
#include "TtrEnclosure.h"
#include "TtrTagDef.h"
#include "TtrTagManager.h"

// ---------------------------------------------------
// Class RuleContext
// ---------------------------------------------------

class RuleContext
{
public:
  RuleContext( TtrText &text,
               TtrFormatSetting &setting );

  void change( SL nCurLineNo );

  /*! Returns the current line filtered.
   */
  const TtrTextLine &line() const;

  /*! Returns the previous line filtered.
   */
  const TtrTextLine &linePrev() const;

  /*! Returns the current line in the text.
   */
  TtrTextLine &line();

  /*! Returns the previous line in the text.
   */
  TtrTextLine &linePrev();

  TtrFormatSetting &setting() const;

  void setLine( EHolderPtrT<TtrTextLine> &rhptrLine );
  void setLinePrev( EHolderPtrT<TtrTextLine> &rhptrLinePrev );

private:  
  TtrText &_text;
  EHolderPtrT<TtrTextLine> _hptrLine;
  EHolderPtrT<TtrTextLine> _hptrLinePrev;
  TtrTextLine *_pLine;
  TtrTextLine *_pLinePrev;
  TtrFormatSetting &_setting;
  SL _nCurLineNo;
};


RuleContext::RuleContext( TtrText &text,
                          TtrFormatSetting &setting ) :
    _text( text ),
    _nCurLineNo( -1 ),
    _setting( setting ),
    _pLine( NULL ),
    _pLinePrev( NULL )
{
}


void 
RuleContext::change( SL nCurLineNo )
{
  if ( _nCurLineNo -1 == nCurLineNo )
  {
    _hptrLinePrev = _hptrLine;
    _pLinePrev = _pLine;
  }
  else
  {
    _pLinePrev = &_text.line( nCurLineNo -1 );
    _hptrLinePrev = EHolderPtrT<TtrTextLine>( _pLinePrev, false ); 
  }

  _pLine = &_text.line( nCurLineNo );
  _hptrLine = EHolderPtrT<TtrTextLine>( _pLine, false );
  _nCurLineNo = nCurLineNo;
}


void 
RuleContext::setLine( EHolderPtrT<TtrTextLine> &rhptrLine )
{
  if ( rhptrLine.get() )
  _hptrLine = rhptrLine;
}


void 
RuleContext::setLinePrev( EHolderPtrT<TtrTextLine> &rhptrLinePrev )
{
  if ( rhptrLinePrev.get() )
    _hptrLinePrev = rhptrLinePrev;
}


TtrTextLine &
RuleContext::line()
{
  return *_pLine;
}


TtrTextLine &
RuleContext::linePrev()
{
  return *_pLinePrev;
}


TtrFormatSetting &
RuleContext::setting() const
{
  return _setting;
}


const TtrTextLine &
RuleContext::line() const
{
  return *_hptrLine.get();
}


const TtrTextLine &
RuleContext::linePrev() const
{
  return *_hptrLinePrev.get();
}

// ---------------------------------------------------
// Class RuleIndentation
// ---------------------------------------------------
class RuleIndentation
{
public:
  RuleIndentation() {}
  virtual ~RuleIndentation() {}

  /*!
   * Applies the rule on the specified context.
   * @param line Current line.
   * @param linePrev Previous line.
   * @param mode Analysis settings.
   * @return \c true if the line must be indented,
   *         otherwise returns \c false.
   */
  virtual bool applyRule( const RuleContext &ctx ) =0;
};


// ---------------------------------------------------
// Class RuleIndentationComposite
// ---------------------------------------------------
class RuleIndentationComposite : public RuleIndentation
{
public:
  RuleIndentationComposite() {}
  virtual ~RuleIndentationComposite() {}
  
  /*!
   * Applies the rule on the specified context.
   * Calls the applyRule() method on each added rule. If one of the rule returns
   * \c true then return \c true.
   * @param line Current line.
   * @param linePrev Previous line.
   * @param mode Analysis settings.
   * @return \c true if one of the rule that has been added return \c true.
   */
  virtual bool applyRule( const RuleContext &ctx );

  /*!
   * Adds a rule to the composite.
   * @param pRule Pointer on the rule to add. It will be destroyed by the compisite.
   */
  void add( RuleIndentation *pRule );

  /*! Removes a rule from the composite.
   * @param pRule Pointer on the rule to remove.
   */
  void remove( RuleIndentation *pRule );

private:
  typedef EPtrDynArrayT<RuleIndentation *> ArrayRule;
  ArrayRule _apRule;
};


// ---------------------------------------------------
// Class RuleIndentationIsBlank
// ---------------------------------------------------
class RuleIndentationIsBlank : public RuleIndentation
{
public:
  RuleIndentationIsBlank() {}
  virtual ~RuleIndentationIsBlank() {}

  virtual bool applyRule( const RuleContext &ctx  )
  {
    return ctx.line().isBlank();
  }
};


// ---------------------------------------------------
// Class RuleIndentationIsBlankPrev
// ---------------------------------------------------
class RuleIndentationIsBlankPrev : public RuleIndentation
{
public:
  RuleIndentationIsBlankPrev() {}
  virtual ~RuleIndentationIsBlankPrev() {}

  virtual bool applyRule( const RuleContext &ctx  )
  {
    return ctx.linePrev().isBlank();
  }
};


// ---------------------------------------------------
// Class RuleIndentationIsIndented
// ---------------------------------------------------
class RuleIndentationIsIndented : public RuleIndentation
{
public:
  RuleIndentationIsIndented( SL nLeftMargin ) :
      _nLeftMargin( nLeftMargin ) {}
  virtual ~RuleIndentationIsIndented() {}

  virtual bool applyRule( const RuleContext &ctx  )
  {
    return ctx.line().indentCount() > _nLeftMargin;
  }

private:
  SL _nLeftMargin;
};


// ---------------------------------------------------
// Class RuleIndentationBeginWith
// ---------------------------------------------------
class RuleIndentationBeginWith : public RuleIndentation
{
public:
  RuleIndentationBeginWith( const EString &strBeginChar ) :
      _strBeginChar( strBeginChar )
  {
  }

  virtual ~RuleIndentationBeginWith() {}

  virtual bool applyRule( const RuleContext &ctx  )
  {
    return _strBeginChar.contains( ctx.line().firstNonWhiteChar() );
  }

private:
  EString _strBeginChar;
};


// ---------------------------------------------------
// Class RuleIndentationIsScript
// ---------------------------------------------------
class RuleIndentationIsScript : public RuleIndentation
{
public:
  RuleIndentationIsScript() {}
  virtual ~RuleIndentationIsScript() {}

  virtual bool applyRule(  const RuleContext &ctx  )
  {
    if ( ctx.line().contains( _T( ':' ) ) )
    { // check if ':' is the last character on the line, or if there are more than 1
      return ( ctx.line().at( ctx.line().firstTrailingSpacePos() -1 ) != _T( ':' ) )  ||
             ( ctx.line().pos( _T( ':' ) ) != ctx.line().posLast( _T( ':' ) ) );
    }

    return false;
  }
};


// ---------------------------------------------------
// Class RuleIndentationIsSeparator
// ---------------------------------------------------
class RuleIndentationIsSeparator : public RuleIndentation
{
public:
  RuleIndentationIsSeparator() {}
  virtual ~RuleIndentationIsSeparator() {}

  virtual bool applyRule(  const RuleContext &ctx  )
  {
    return ctx.line().isSeparatorLine( ctx.setting().getFormatSeparatorChar() )  ||
           ctx.linePrev().isSeparatorLine( ctx.setting().getFormatSeparatorChar() );
  }
};


// ---------------------------------------------------
// Class RuleIndentationIsEosPrev
// ---------------------------------------------------
class RuleIndentationIsEosPrev : public RuleIndentation
{
public:
  RuleIndentationIsEosPrev() {}
  virtual ~RuleIndentationIsEosPrev() {}

  virtual bool applyRule( const RuleContext &ctx  )
  {
    SL nPosLastChar = ctx.linePrev().firstTrailingSpacePos() -1;
    return ( nPosLastChar >= 0 )  &&
           ctx.setting().getFormatModeEosChar().contains( 
              ctx.linePrev().at( nPosLastChar ) );
  }
};


// ---------------------------------------------------
// Class RuleIndentationRightMargin
// ---------------------------------------------------

class RuleIndentationRightMargin : public RuleIndentation
{
public:
  RuleIndentationRightMargin( SL nRightMargin,
                              bool bIndentBlankLine ) :
      _nRightMargin( nRightMargin ),
      _bIndentBlankLine( bIndentBlankLine )
  {
    initWordLetter();
  }

  virtual ~RuleIndentationRightMargin() {}

  virtual bool applyRule( const RuleContext &ctx  )
  {
    SL nFirstWordLength = findFirstWordLength( ctx );
    if ( nFirstWordLength == 0 )    // blank line ? 
      return _bIndentBlankLine;         // Returns "indent blank line" rule status.

    // The word can fit on the previous line?
    return (ctx.linePrev().length() + nFirstWordLength +1 ) <= _nRightMargin;
  }

  static SL findFirstWordLength( const RuleContext &ctx )
  {
    EString strWord;
    SL nWordStart = ctx.line().countHeadingSpace();
    SL nWordEnd;
    for ( nWordEnd = nWordStart; nWordEnd < ctx.line().length(); ++nWordEnd )
      if ( !_vbIsWordLetter[ ctx.line().at( nWordEnd ) ] )
        break;

    return nWordEnd - nWordStart;
  }

  void initWordLetter()
  {
    if (_bWordLetterInitialized )
      return;

    int nIdx;
//    const EString strWordLetter( 
//        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
//        "0123456789':,;" 
//        );
    for ( nIdx =0; nIdx < 256; ++nIdx )
      _vbIsWordLetter[ nIdx ] = nIdx > 32;
//      _vbIsWordLetter[ nIdx ] = strWordLetter.contains( nIdx );

    _bWordLetterInitialized = false;
  }

private:
  static bool _vbIsWordLetter[256];
  static bool _bWordLetterInitialized;
  int _nRightMargin;
  bool _bIndentBlankLine; 
};

bool RuleIndentationRightMargin::_vbIsWordLetter[256];
bool RuleIndentationRightMargin::_bWordLetterInitialized =false;

// Implementation class RuleIndentationComposite
// ---------------------------------------------------

bool
RuleIndentationComposite::applyRule( const RuleContext &ctx  )
{
  ArrayRule::EIterator it( _apRule );
  for ( it.begin(); !it.isEnd(); it.next() )
    if ( it.item()->applyRule( ctx ) )
      return true;
  return false;
}


void 
RuleIndentationComposite::add( RuleIndentation *pRule )
{
  if ( pRule )
    _apRule.append( pRule );
}


void 
RuleIndentationComposite::remove( RuleIndentation *pRule )
{
  if ( pRule )
  {
    SL nIdx = _apRule.find( pRule );
    if ( nIdx >= 0)
      _apRule.remove( nIdx );
  }
}


// ---------------------------------------------------
// Class RuleIndentationIndentStringChange
// ---------------------------------------------------
class RuleIndentationIndentStringChange : public RuleIndentation
{
public:
  RuleIndentationIndentStringChange() {}
  virtual ~RuleIndentationIndentStringChange() {}

  virtual bool applyRule( const RuleContext &ctx  )
  {
    RuleContext &ctxReal = (RuleContext &)ctx;
    return ( ctxReal.linePrev().indentLength() != ctxReal.line().indentLength() )  
           &&  (ctxReal.line().indentLength() > 0);
  }
};


// ---------------------------------------------------
// Class RuleEnclosed
// ---------------------------------------------------
class RuleEnclosed
{
public:
  RuleEnclosed() {}
  virtual ~RuleEnclosed() {}

  virtual void processLine( const RuleContext &ctx ) =0;

  virtual bool isInside() const =0;
};


// ---------------------------------------------------
// Class RuleEnclosedComposite
// ---------------------------------------------------
class RuleEnclosedComposite : public RuleEnclosed
{
public:
  RuleEnclosedComposite() {}
  virtual ~RuleEnclosedComposite() {}

  virtual void processLine( const RuleContext &ctx );

  virtual bool isInside() const;

  void add( RuleEnclosed *pRule );
  void remove( RuleEnclosed *pRule );

private:
  typedef EPtrDynArrayT<RuleEnclosed *> ArrayRule;
  ArrayRule _apRule;
};



void
RuleEnclosedComposite::processLine( const RuleContext &ctx  )
{
  ArrayRule::EIterator it( _apRule );
  for ( it.begin(); !it.isEnd(); it.next() )
    it.item()->processLine( ctx );
}


bool 
RuleEnclosedComposite::isInside() const
{
  ArrayRule::EIterator it( _apRule );
  for ( it.begin(); !it.isEnd(); it.next() )
    if ( it.item()->isInside() )
      return true;
  return false;
}


void 
RuleEnclosedComposite::add( RuleEnclosed *pRule )
{
  if ( pRule )
    _apRule.append( pRule );
}


void 
RuleEnclosedComposite::remove( RuleEnclosed *pRule )
{
  if ( pRule )
  {
    SL nIdx = _apRule.find( pRule );
    if ( nIdx >= 0)
      _apRule.remove( nIdx );
  }
}


// ---------------------------------------------------
// Class RuleEnclosedString
// ---------------------------------------------------
class RuleEnclosedString : public RuleEnclosed
{
public:
  RuleEnclosedString( bool bIsInside,
                      const TtrEnclosure &enclosure ) :
      _bIsInside( bIsInside ),
      _enclosure( enclosure ) {}
  virtual ~RuleEnclosedString() {}

  virtual void processLine( const RuleContext &ctx )
  {
    SL nIdx =0;
    SL nLength = ctx.linePrev().length();

    while ( nIdx < nLength )
    {
      const EString &strToMatch = _bIsInside ? _enclosure.getClosure() : 
                                               _enclosure.getOpening();
      nIdx = ctx.linePrev().pos( nIdx, strToMatch );
      if ( nIdx >= 0 )
      {
        _bIsInside = !_bIsInside;
        ++nIdx;
      }
      else
        break;
    }
  }

  virtual bool isInside() const
  {
    return _bIsInside;
  }
private:
  bool _bIsInside;
  const TtrEnclosure &_enclosure;
};


// ---------------------------------------------------
// Class RuleEnclosed
// ---------------------------------------------------
/*! \brief Rule to prevent the indentation if the line starts with a lower case.
 */
class RuleEnclosedNotLowerCaseLine : public RuleEnclosed
{
public:
  RuleEnclosedNotLowerCaseLine() :
      _bFirstIsLower( false ) {}
  virtual ~RuleEnclosedNotLowerCaseLine() {}

  virtual void processLine( const RuleContext &ctx );

  virtual bool isInside() const;
private:
  bool _bFirstIsLower;
};



void 
RuleEnclosedNotLowerCaseLine::processLine( const RuleContext &ctx )
{
  _bFirstIsLower = ::islower( ctx.line().firstNonWhiteChar() )  &&
                    ctx.line().firstNonWhiteChar() != 0;
}


bool 
RuleEnclosedNotLowerCaseLine::isInside() const
{
  return _bFirstIsLower;
}


// ---------------------------------------------------
// ---------------------------------------------------
// Class TtrTextFormaterV0
// ---------------------------------------------------
// ---------------------------------------------------


void 
TtrTextFormaterV0::doFormatBlock( TtrTextBlock &block,
                                  TtrFormatSetting &setting )
{
  TtrText &text = block.text();
  const TtrTextBlockDef &range = block.range();

  if ( !block.isValid() )
    return;

  text.setModified();

  setProgressRange( range.firstLine(),
                    range.lastLine() );
  startProgress();

  RuleIndentationComposite ruleIndentation;
  setupIndentationRules( setting, ruleIndentation );

  RuleEnclosedComposite ruleEnclosed;
  setupEnclosedRules( setting, ruleEnclosed );

  // Specific for the first line
  SL nLineNo = range.firstLine();
//  text.line( nLineNo ).setIndentLevel( 
//      ruleEnclosed.isInside( RuleContext( text, nLineNo, mode ) ) ? -1 : 1 );
  text.line( nLineNo ).setIndentLevel( 1 );
  freeEngineTag( text.line( nLineNo ) );

  // For the other lines
  RuleContext ctx( text, setting );
  for ( ++nLineNo; nLineNo <= range.lastLine(); ++nLineNo )
  {
    TtrTextLine &line = text.line( nLineNo );

    if ( (nLineNo & 127) == 0 )  // show progress in status bar
      setProgress( nLineNo );

    ctx.change( nLineNo );      // Moves the context to the current line.
    detectIndentString( ctx );

    freeEngineTag( line );

    ruleEnclosed.processLine( ctx );
    if ( !ruleEnclosed.isInside()  &&  
         ruleIndentation.applyRule( ctx ) )
    {
      line.setIndentLevel( 1 );   // set flag
      applyCenteringRule( setting, ctx );
    }
    else
      line.setIndentLevel( -1 );

  }

  stopProgress();
}


void 
TtrTextFormaterV0::setupIndentationRules( TtrFormatSetting &setting, 
                                          RuleIndentationComposite &ruleIndentation )
{
  if ( setting.getFormatIndentBlankLine() )
    ruleIndentation.add( new RuleIndentationIsBlank() );
  if ( setting.getFormatIndentIndentedLine() )
    ruleIndentation.add( 
        new RuleIndentationIsIndented( 
            setting.getFormatUseLeftMargin() ? setting.getFormatLeftMargin() : 0 ) );
  if ( setting.getFormatIndentContainDoubleDot() )
    ruleIndentation.add( new RuleIndentationIsScript() );
  if ( setting.getFormatIndentSeparator() )
    ruleIndentation.add( new RuleIndentationIsSeparator() );
  if ( setting.getFormatIndentPreviousLineIsBlank() )
    ruleIndentation.add( new RuleIndentationIsBlankPrev() );
  if ( setting.getFormatModeEos() )
    ruleIndentation.add( new RuleIndentationIsEosPrev() );
  
  EString strBeginChar;
  if ( setting.getFormatIndentBeginWith() )
    strBeginChar.append( setting.getFormatIndentBeginWithChar() );

  if ( !strBeginChar.isNull() )
    ruleIndentation.add( new RuleIndentationBeginWith( strBeginChar ) );

  if ( setting.getFormatUseRightMargin() )
    ruleIndentation.add( 
        new RuleIndentationRightMargin( setting.getFormatRightMargin(),
                                        setting.getFormatIndentBlankLine() ) );

  if ( setting.getNeedFilterIndentString() )
    ruleIndentation.add( new RuleIndentationIndentStringChange() );
}


void 
TtrTextFormaterV0::setupEnclosedRules( TtrFormatSetting &setting, 
                                       RuleEnclosedComposite &ruleEnclosed )
{
  if ( setting.getFormatUseEnclosure() )
  {
    EAutoPtrT<ItTtrEnclosure> pItEnclosure( setting.getFormatEnclosureList().iterator() );
    for ( pItEnclosure->begin(); !pItEnclosure->isEnd(); pItEnclosure->next() )
    {
      ruleEnclosed.add( new RuleEnclosedString( false, *pItEnclosure->item() ) );
    }
  }

  if ( setting.getFormatNotLowerCaseLine() )
    ruleEnclosed.add( new RuleEnclosedNotLowerCaseLine() );
}


void 
TtrTextFormaterV0::freeEngineTag( TtrTextLine &line )
{
  TtrTagList &list = line.tagList();
  for ( SL nIdx =0; nIdx < list.tagCount(); )
  {
    const TtrTagRef &refTag = list.tagAt( nIdx );
    if ( refTag->isEngineDefined() )
    {
      list.removeTag( refTag );
    }
    else
      ++nIdx;
  }
}


void 
TtrTextFormaterV0::applyCenteringRule( TtrFormatSetting &setting,
                                       RuleContext &ctx )
{
  const TtrTextLine &line = ctx.line();

  if ( !setting.getFormatIndentSeparator()  ||
       !line.isSeparatorLine( ctx.setting().getFormatSeparatorChar() ) )
    return;

  ctx.line().tagList().addTag( 
      TtrTagManager::manager().getTag( TtrTagDef::sidTagCenter,
                                       TtrTagAttListRef::nullObject,
                                       true ) );
}


void 
TtrTextFormaterV0::detectIndentString( RuleContext &ctx )
{
  if ( !ctx.setting().getNeedFilterIndentString() )
    return;

  detectIndentString( ctx, ctx.linePrev() );
  detectIndentString( ctx, ctx.line() );

  if ( ctx.linePrev().indentLength() > 0 )
  {
    EHolderPtrT<TtrTextLine> hptrLinePrev( 
      new TtrTextLine( 
          ctx.linePrev().mid( ctx.linePrev().indentLength(), -1 ) ) );
    ctx.setLinePrev( hptrLinePrev );
  }

  if ( ctx.line().indentLength() > 0 )
  {
    EHolderPtrT<TtrTextLine> hptrLine( 
      new TtrTextLine( 
          ctx.line().mid( ctx.line().indentLength(), -1 ) ) );
    ctx.setLine( hptrLine );
  }
}


void 
TtrTextFormaterV0::detectIndentString( RuleContext &ctx,
                                       TtrTextLine &line )
{
  static const EString strIndentChar( ">:" );

  SL nIdx;
  bool bWasSpace = false;
  for ( nIdx =0; nIdx < line.length(); ++nIdx )
  {
    char cThis = line.at( nIdx );
    if ( strIndentChar.contains( cThis ) )
    {
      bWasSpace = false;
      continue;
    }

    if ( cThis == ' ' )
    {
      if ( bWasSpace )    // not two spaces one after the other.
        break;
      bWasSpace = true;
      continue;
    };
    
    break;
  }

  if ( nIdx == 1  &&  line.at( 0 ) == ' ' )
    nIdx = 0;

  line.setIndentLength( nIdx );
}
