// //////////////////////////////////////////////////////////////////////////
// Implementation file TtrWriterHtml.cpp for class TtrWriterHtml
// (c)Copyright 2000, Baptiste Lepilleur.
// Created: 2000/09/17
// //////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "TtrWriterHtml.h"
#include "TtrWriterSetting.h"
#include "TtrFmtBlank.h"
#include "TtrFmtPara.h"
#include "TtrTagDef.h"


TtrWriterHtml::TtrWriterHtml( ETextStream &rstreamOut,
                              const TtrWriterSetting &setting ) :
    _rstreamOut( rstreamOut ),
    _nIndentLevel( 0 )
{
  setup( setting );
}


TtrWriterHtml::~TtrWriterHtml()
{
}


void 
TtrWriterHtml::visitBlank( TtrFmtBlank &blank )
{
  SL nLineCount = blank.lineCount() + 
                  ( _bNeedLineBreak ? 1 : 0 );

  if ( nLineCount == 1 )
    stream()  << "<BR>"  <<  endl;
  else
  {
    stream()  <<  "<P>"  <<  endl;
    while ( --nLineCount > 1 )
      stream()  <<  "<BR>"  <<  endl;
  }
  
  _bNeedLineBreak = false;
}


void 
TtrWriterHtml::visitPara( TtrFmtPara &para )
{
  flushPendingLineBreak();

  EString strEscapedContent = escape( para.content() );
  
  changeIndentLevelTo( para.indent().indentString().length() -
                       para.indent().indentString().count( ' ' ) );

  if ( para.tagList().findTag( TtrTagDef::sidTagCenter ).isNull() )
  {       // indent the paragraph
    if ( _bKeepOriginalIndent )   // with the original indent
    {
      stream()  <<  makeNbspace( para.indent().indentString() )
                <<  strEscapedContent;
    }
    else    // with the specified indent
    {
      EString strOpen;
      strOpen.append( _strIndentStepOpen, para.indent().indentLevel() );

      EString strClose;
      strClose.append( _strIndentStepClose, para.indent().indentLevel() );

      stream()  <<  strOpen  <<  strEscapedContent  <<  strClose;
    }
    _bNeedLineBreak = true;
  }
  else    // center the paragraph.
  {
    // note: the CENTER tag automatically jump one line after closure.
    stream()  <<  "<CENTER>"  <<  strEscapedContent  << "</CENTER>"  <<  endl;
  }
}


void 
TtrWriterHtml::changeIndentLevelTo( SL nLevel )
{
  static const char *acszColor[] = {
    "Navy",
    "Blue",
    "Teal",
    "Aqua",
    "Fuchsia",
    "Purple",
    "Red",
    "Maroon",
    "Green",
    "Lime",
    "Olive",
    "Yellow" };
  const SL nColorCount = sizeof( acszColor ) / sizeof( const char * );

  if ( nLevel < 0 )
    nLevel =0;

  while ( nLevel != _nIndentLevel )
  {
    if ( nLevel < _nIndentLevel )   // lower the indent level
    {
      stream()  <<  endl  <<  "</TD></TR></TABLE>"  <<  endl;
      --_nIndentLevel;
    }
    else    //  raise the indent level
    {
      EString strColor( acszColor[ _nIndentLevel % nColorCount ] );
      stream()  <<  "<TABLE WIDTH=\"100%\" BORDER=0 COLS=2><TR>"
                <<  "<TD WIDTH=5 BGCOLOR="
                <<  strColor  << ">&nbsp;</TD><TD>"  <<  endl;
      ++_nIndentLevel;
    }
  }
}


void 
TtrWriterHtml::start()
{
  stream()  <<  "<HTML><BODY>"  <<  endl;
  _nIndentLevel = 0;
}


void 
TtrWriterHtml::end()
{
  flushPendingLineBreak();
  changeIndentLevelTo( 0 );
  stream()  <<  "</BODY></HTML>"  <<  endl;
}


void 
TtrWriterHtml::setup( const TtrWriterSetting &setting )
{
  _bKeepOriginalIndent = ( setting.getWriterLineMode() == wslmIndentAsOriginal );
  
  _strIndentStepOpen.setNull();
  _strIndentStepClose.setNull();
  if ( !_bKeepOriginalIndent )
  {
    // Generates the following indents string...
    //  "&nbsp;&nbsp;&nbsp; "
    //  "<DL><DD></DL>"  --> replaced with "&nbsp;&nbsp;&nbsp; "
    //                       because it was causing a line jump...
    SL nIndentSize = setting.getWriterReindentSize();
    if ( setting.getWriterLineMode() == wslmIndentWithTab )
    {
//      _strIndentStepOpen.append( EString( "<DL><DD>" ), nIndentSize );
//      _strIndentStepClose.append( EString( "</DL>" ), nIndentSize );
      _strIndentStepOpen.append( EString( "&nbsp;" ), (nIndentSize *4) -1 );
      _strIndentStepOpen.append( " " );
    }
    else
    {
      _strIndentStepOpen.append( EString( "&nbsp;" ), nIndentSize -1 );
      _strIndentStepOpen += ' ';
    }
  }


  _bNeedLineBreak = false;
}


void 
TtrWriterHtml::flushPendingLineBreak()
{
  if ( _bNeedLineBreak )
  {
    stream()  <<  "<BR>"  <<  endl;
    _bNeedLineBreak = false;
  }
}




EString 
TtrWriterHtml::escape( const EString &strText )
{
  EString str;
  EString strChar;
  SL nIdx;
  for ( nIdx =0; nIdx < strText.length(); ++nIdx )
  {
    TCHAR tcChar = strText[ nIdx ];
    
    switch ( tcChar )
    {
    case '<': strChar = "&lt;"; break;
    case '>': strChar = "&gt;"; break;
    case '': strChar = "&eacute;"; break;
    case '': strChar = "&egrave;"; break;
    case '': strChar = "&agrave;"; break;
    case '': strChar = "&ugrave;"; break;
    default: strChar = tcChar; break;
    }

    str += strChar;
  }

  return str;
}


EString 
TtrWriterHtml::makeNbspace( const EString &strText )
{
  EString str;
  EString strChar;
  SL nIdx;
  SL nLength = strText.length();
  SL nLast = nLength -1;
  for ( nIdx =0; nIdx < nLength; ++nIdx )
  {
    TCHAR tcChar = strText[ nIdx ];
    
    switch ( tcChar )
    {
    case '\t': 
      {
        if ( nIdx < nLast  &&  strText[nIdx+1] == ' ' )
          strChar = "&nbsp;&nbsp;&nbsp;&nbsp;";
        else
          strChar = "&nbsp;&nbsp;&nbsp; ";
        break;
      }
    case ' ': 
      {
        if ( nIdx < nLast  &&  strText[nIdx+1] == ' ' )
          strChar = "&nbsp;";
        else
          strChar = " ";
        break;
      }
    default: strChar = tcChar; break;
    }

    str += strChar;
  }

  return str;
}

